""" Misc classes and functions. """

import datetime, inspect, json, logging, os, shutil, socket, time
from enum import Enum
from hashlib import sha256
import __main__, random_user_agent.params, random_user_agent.user_agent, requests

class Config:
	""" Config class. """

	@classmethod
	def load(cls, filename: str=None, defaults: dict=None) -> dict:
		""" Loads config file.

				Args:
					filename:			Filename, or default filename (config.txt).
					defaults:			Dict of default values to be used for any not indicated in file.

				Returns:
					arg:					Config dict, or copy of default dict (or empty dict) on failure.
		"""

		return_dict = (defaults or {}).copy()
		try:
			filename = filename or cls.get_filename()
			if os.path.isfile(filename) is False:
				Log.info(f'{filename} does not exist - using defaults')
			else:
				with open(filename, 'rt', encoding='UTF-8') as f:
					config = json.loads(f.read().strip())
					return_dict.update(config)
		except Exception as e:
			Log.error(str(e))
		return return_dict

	@classmethod
	def save(cls, config: dict=None, filename: str=None):
		""" Saves config dict.

				Args:
					filename:			Filename, or default filename (config.txt).
					defaults:			Dict of default values to be used for any not indicated in file.
		"""

		filename = filename or cls.get_filename()
		with open(filename, 'wt', encoding='UTF-8') as f:
			f.write(json.dumps(config or {}))

	@staticmethod
	def get_filename() -> str:
		""" Gets default filename. """

		if hasattr(__main__, '__file__'):
			folder = os.path.dirname(os.path.realpath(__main__.__file__))
		else:		# the above instruction fails for spawned processes with no file.
			folder = os.getcwd()
		return os.path.join(folder, 'config.txt')

class Log:
	""" Log class using Python logging module. """

	ERROR = logging.ERROR
	WARNING = logging.WARNING
	INFO = logging.INFO
	DEBUG = logging.DEBUG

	loggers = {}
	file_handlers = {}
	console_handler = None
	file_levels = {}
	console_level = logging.WARNING

	format_string = '%(asctime)s\t%(class_name)s\t%(function_name)s\t%(levelname)s\t%(message)s'
	formatter = logging.Formatter(format_string, datefmt='%Y%m%d %H:%M:%S')

	@classmethod
	def parse(cls, record: str) -> dict|None:
		""" Parses log message string into fields. """

		record = record.strip()
		try:
			fields = record.split('\t')
			if len(fields) != 5:
				return None
			dt = datetime.datetime.strptime(fields[0].strip(), '%Y%m%d %H:%M:%S')
			class_name = fields[1].strip()
			method_name = fields[2].strip()
			level = fields[3].strip()
			message = fields[4].strip()
			return {'dt': dt, 'class': class_name, 'method': method_name, 'level': level, 'message': message}
		except Exception as e:
			Log.error(f'Could not parse {record}: {e}')
			return None

	@classmethod
	def get_logger(cls, filename: str=None) -> logging.Logger:
		""" Gets or creates a logger for the specified file. """

		filename = filename or cls.get_filename()
		if filename in cls.loggers:
			return cls.loggers[filename]

		# create logger
		logger = logging.getLogger(filename)
		logger.setLevel(cls.DEBUG)		# catch all errors, but filter for output
		logger.propagate = False
		file_handler = logging.FileHandler(filename, mode='a', encoding='UTF-8')
		file_handler.setLevel(cls.file_levels.get(filename, cls.INFO))
		file_handler.setFormatter(cls.formatter)
		logger.addHandler(file_handler)
		cls.loggers[filename] = logger
		cls.file_handlers[filename] = file_handler

		# create console handler if it does not yet exist
		if cls.console_handler is None:
			cls.console_handler = logging.StreamHandler()
			cls.console_handler.setFormatter(cls.formatter)
		cls.console_handler.setLevel(cls.console_level)
		logger.addHandler(cls.console_handler)

		return cls.loggers[filename]

	@classmethod
	def get_caller_info(cls, stack_depth: int=3) -> (str|None, str|None):
		""" Extracts caller class and function names from stack.

		Args:
			stack_depth:				Number of frames to go back in the stack

		Returns:
			param1:							Class name.
			param2:							Function name.
		"""
		frame = inspect.currentframe()
		try:
			for _ in range(stack_depth):
				if frame is not None:
					frame = frame.f_back
			if frame is not None:
				function_name = frame.f_code.co_name
				class_name = ''
				if 'self' in frame.f_locals:
					class_name = frame.f_locals['self'].__class__.__name__
				elif 'cls' in frame.f_locals:
					class_name = frame.f_locals['cls'].__name__
				# check if this is a static method of a class
				elif 'self' not in frame.f_locals and 'cls' not in frame.f_locals:
					# infer from module and qualname
					if hasattr(frame.f_code, 'co_qualname'):
						qualname = frame.f_code.co_qualname
						if '.' in qualname:
							class_name = qualname.split('.')[0]
				# if still no class name, use module name
				if not class_name:
					module = inspect.getmodule(frame)
					if module:
						class_name = module.__name__.split('.')[-1]
					else:
						class_name = None
				return class_name, function_name
		finally:
			del frame

		return None, None

	@classmethod
	def reset(cls, filename: str=None):
		""" Resets log at specified filename or default filename (log.txt). """

		filename = filename or cls.get_filename()
		if filename in cls.loggers:
			logger = cls.loggers[filename]
			if filename in cls.file_handlers:
				handler = cls.file_handlers[filename]
				handler.close()
				logger.removeHandler(handler)
				del cls.file_handlers[filename]
			del cls.loggers[filename]
		if os.path.isfile(filename):
			os.unlink(filename)
		cls.info('Starting new log', filename=filename)

	@classmethod
	def backup(cls, filename: str=None, backup_filename: str=None, backup_folder: str=None) -> bool:
		""" Saves current log in backup folder.

		Args:
			filename: 				Filename, or default filename (log.txt).
			backup_filename:  Filename of backup file, or default log_(YYYYmmdd_HHMMSS).txt.
			backup_folder: 		Folder for backups, or "logs" folder in same folder as log.

		Returns:
			param1:						Success indicator.
		"""

		try:
			filename = filename or cls.get_filename()
			if not os.path.isfile(filename):  # no log to backup
				return True

			backup_folder = backup_folder or os.path.join(os.path.dirname(filename), 'logs')
			if not os.path.isdir(backup_folder):
				os.makedirs(backup_folder, exist_ok=True)
				if not os.path.isdir(backup_folder):
					cls.warning(f'{backup_folder} does not exist and could not be created')
					return False
			if filename in cls.file_handlers:
				handler = cls.file_handlers[filename]
				handler.close()
				cls.loggers[filename].removeHandler(handler)
				del cls.file_handlers[filename]
				del cls.loggers[filename]

			if backup_filename is None:
				current_time = datetime.datetime.now().strftime('%Y%m%d_%H%M%S.%f')
				backup_filename = f'log_{current_time}.txt'
			backup_filename = os.path.join(backup_folder, backup_filename)
			shutil.move(filename, backup_filename)

			cls.info(f'Backed up previous log ({backup_filename}) and starting new log')
			return True

		except Exception as e:
			cls.error(str(e))
			return False

	@classmethod
	def set_file_level(cls, level: int, filename: str=None):
		""" Set the logging level for a specific file. """

		filename = filename or cls.get_filename()
		cls.file_levels[filename] = level
		if filename in cls.file_handlers:
			cls.file_handlers[filename].setLevel(level)

	@classmethod
	def set_console_level(cls, level: int):
		""" Set the console logging level. """

		cls.console_level = level
		if cls.console_handler is not None:
			cls.console_handler.setLevel(level)

	@classmethod
	def error(cls, message, filename: str=None):
		""" Writes error message to log. """

		class_name, function_name = cls.get_caller_info()
		logger = cls.get_logger(filename)
		logger.error(str(message), extra={'class_name': class_name, 'function_name': function_name})

	@classmethod
	def warning(cls, message, filename: str=None):
		""" Writes warning message to log. """

		class_name, function_name = cls.get_caller_info()
		logger = cls.get_logger(filename)
		logger.warning(str(message), extra={'class_name': class_name, 'function_name': function_name})

	@classmethod
	def info(cls, message, filename: str=None):
		""" Writes info message to log. """

		class_name, function_name = cls.get_caller_info()
		logger = cls.get_logger(filename)
		logger.info(str(message), extra={'class_name': class_name, 'function_name': function_name})

	@classmethod
	def debug(cls, message, filename: str=None):
		""" Writes debug message to log. """
		class_name, function_name = cls.get_caller_info()

		logger = cls.get_logger(filename)
		logger.debug(str(message), extra={'class_name': class_name, 'function_name': function_name})

	@classmethod
	def read(cls, filename: str=None) -> list:
		""" Read log at specified filename. """

		filename = filename or cls.get_filename()
		if not os.path.isfile(filename):
			return []
		if filename in cls.file_handlers:
			cls.file_handlers[filename].flush()
		with open(filename, 'rt', encoding='UTF-8') as f:
			log_lines = [line.strip() for line in f.readlines()]
		return log_lines

	@classmethod
	def get_modification_date(cls, filename: str=None):
		""" Get file modification date of specified filename. """

		filename = filename or cls.get_filename()
		return None if not os.path.isfile(filename) else os.path.getmtime(filename)

	@staticmethod
	def get_filename() -> str:
		""" Returns default filename. """

		if hasattr(__main__, '__file__'):
			folder = os.path.dirname(os.path.realpath(__main__.__file__))
		else:		 	# the above instruction fails for spawned processes with no file
			folder = os.getcwd()
		return os.path.join(folder, 'log.txt')

class Color(Enum):
	""" Color constants. """

	BLACK = (0, 0, 0)
	WHITE = (255, 255, 255)
	GRAY = (127, 127, 127)
	DARK_GRAY = (64, 64, 64)
	LIGHT_GRAY = (196, 196, 196)
	RED = (255, 0, 0)
	DARK_RED = (127, 0, 0)
	LIGHT_RED = (255, 64, 64)
	GREEN = (0, 255, 0)
	DARK_GREEN = (0, 127, 0)
	LIGHT_GREEN = (64, 255, 64)
	BLUE = (0, 0, 255)
	DARK_BLUE = (0, 0, 127)
	LIGHT_BLUE = (64, 64, 255)
	ORANGE = (255, 165, 0)
	PINK = (255, 192, 203)
	PURPLE = (128, 0, 128)
	CYAN = (0, 255, 255)
	MAGENTA = (255, 0, 255)
	YELLOW = (238, 210, 2)
	BROWN = (101, 67, 33)

class Misc:
	""" Misc class for misc functions. """

	key_names = ('backspace', 'tab', 'return', 'escape', 'space', '!', '"', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '_', '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'delete', 'f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7', 'f8', 'f9', 'f10', 'f11', 'f12', 'insert' ,'home', 'end', 'right', 'left', 'up', 'down', 'left ctrl', 'left shift', 'right ctrl', 'right shift')
	key_symbols = {k: k for k in 'abcdefghijklmnopqrstuvwxyz`1234567890-=~!@#$%^&*()_+[]\\{}|;\':",./<>?'}
	key_symbols['space'] = ' '
	key_shift_symbols = {'a': 'A', 'b': 'B', 'c': 'C', 'd': 'D', 'e': 'E', 'f': 'F', 'g': 'G', 'h': 'H', 'i': 'I', 'j': 'J', 'k': 'K', 'l': 'L', 'm': 'M', 'n': 'N', 'o': 'O', 'p': 'P', 'q': 'Q', 'r': 'R', 's': 'S', 't': 'T', 'u': 'U', 'v': 'V', 'w': 'W', 'x': 'X', 'y': 'Y', 'z': 'Z', '1': '!', '2': '@', '3': '#', '4': '$', '5': '%', '6': '^', '7': '&', '8': '*', '9': '(', '0': ')', '`': '~', '-': '_', '=': '+', '[': '{', ']': '}', '\\': '|', ';': ':', '\'': '"', ',': '<', '.': '>', '/': '?'}

	null_date = datetime.datetime.utcfromtimestamp(0)
	user_agent = None

	@staticmethod
	def choose_user_agent():
		""" Chooses a pouplar user agent. """

		p=[random_user_agent.params.Popularity.POPULAR.value]
		return random_user_agent.user_agent.UserAgent(popularity=p).get_random_user_agent()

	@staticmethod
	def calculate_hashcode(data: dict) -> str:
		""" Calculates a sha256 hashcode for a JSON object. """

		json_string = json.dumps(data, separators=(',', ':')).encode('utf-8')
		hashcode = sha256(json_string).hexdigest()
		return hashcode

	@classmethod
	def http_request(cls, url: str, method: str=None, headers: dict=None,
		data: dict=None, password: str=None, timeout: int=30,
		randomize_user_agent: bool=False, timestamp: str=None) -> (int|None, str):
		""" Sends an HTTP GET or POST request via requests.

				Args:
					url:					URL.
					method:				GET or POST. Defaults to POST if data is included, or GET if not.
					headers:			Extra headers to specify in request.
					data:					Dictionary of data to transmit as JSON payload in POST request.
					password:			password for signing URL.
					timeout:			timeout in seconds.
					randomize_user_agent:	Randomize user agent. False = use existing popular agent.
					timestamp:		Optional timestamp to append to URL before hashing.

				Returns:
					arg1:					HTTP status code, or None on failure.
					arg2:					Signed URL, or error message on failure.
		"""

		if password is not None:
			result, message = cls.sign_request(url, password, timestamp)
			if result is False:
				return (False, f'Error signing request: {message}')
			url = message
		headers = headers or {}
		if cls.user_agent is None:
			cls.user_agent = cls.choose_user_agent()
		user_agent = cls.choose_user_agent() if randomize_user_agent else cls.user_agent
		status_code = None
		method = 'POST' if data is not None else (method or 'GET')
		try:
			if method.upper() == 'GET':
				headers = {'User-Agent': user_agent}
				response = requests.get(url, headers=headers, timeout=timeout)
				status_code = response.status_code
			else:
				headers = {'User-Agent': user_agent, 'Content-type': 'application/json'}
				response = requests.post(url, headers=headers, json=data, timeout=timeout)
				status_code = response.status_code
		except Exception as e:
			return (None, f'Exception on HTTP request: {e}')
		return (status_code, response)

	@staticmethod
	def sign_request(url: str, password: str, timestamp: int=None) -> (bool, str):
		""" Generates hashcode-signed URLs.

				Args:
					url:					url.
					password:			password for signing URL.
					timestamp:		Optional timestamp to append to URL before hashing.

				Returns:
					arg1:					Success indicator.
					arg2:					Signed URL, or error message on failure.
		"""

		if url is None:
			return (False, 'URL not specified')
		if password is None:
			return (False, 'Password not specified')
		timestamp = int(time.time()) if timestamp is None else int(timestamp)
		url = f'{url}&ts={timestamp}'
		request = password + url
		signature = sha256(request.encode('utf-8')).hexdigest()
		return (True, url + '&hash=' + signature)

	@classmethod
	def get_local_ip_address(cls) -> (bool, str):
		""" Determines local IP address.

				Returns:
					arg1:					Success indicator.
					arg2:					IP address, or error message on failure.
		"""

		with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
			s.settimeout(0)
			try:
				s.connect(('10.254.254.254', 1))			# loopback address
				return (True, s.getsockname()[0])
			except Exception as e:
				return (False, f'Exception: {e}')

	@classmethod
	def test_internet(cls) -> (bool, float|str):
		""" Tests Internet access by querying google.com.

				Returns:
					arg1:					Success indicator.
					arg2:					request duration (in seconds), or error message on failure.
		"""

		start = time.time()
		status_code, response = cls.http_request('https://google.com')
		success = (status_code == 200)
		response = (time.time() - start) if status_code == 200 else response
		return (success, response)

	@staticmethod
	def get_hostname() -> str:
		""" Returns local hostname. """

		return socket.gethostname()

	@staticmethod
	def get_color(color: str|int|tuple, color_scale: float=None) -> tuple|None:
		""" Returns a color.

				Args:
					color:				String (case-insensitive), enum (e.g.: Color.WHITE),
													or custom tuple (e.g.: (255, 255, 0)).
					color_scale:	Optional value to scale color toward black (0.0-1.0).

				Returns:
					arg:					RGB tuple, or None on error.
		"""

		if isinstance(color, tuple):
			color_tuple = color
		elif isinstance(color, str):
			matches = list(c for c in Color if str(c).replace('Color.', '').lower() == color.lower())
			if len(matches) != 1:
				return None
			color_tuple = matches[0].value if len(matches) == 1 else Color.WHITE.value
		else:
			if color not in dir(Color):
				return None
			color_tuple = Color[color].value
		return tuple(int(c * color_scale) for c in color_tuple) if color_scale else color_tuple



""" Jubilee mouse interface class. """

import pygame
from pygame.event import Event
from .base_classes import PointerInterface
from .misc import Log

class MouseInterface(PointerInterface):
	""" Mouse interface class. """

	def handle_event(self, event: Event) -> bool:
		""" Handle mouse events. """

		if event.type == pygame.MOUSEBUTTONDOWN:
			self.down = True
			self.x, self.y = pygame.mouse.get_pos()
			return (self.x is not None and self.y is not None)
		if event.type == pygame.MOUSEBUTTONUP:
			self.down = False
			self.x = None; self.y = None
			return False
		Log.warning(f'Could not handle event of type {event.type}')
		return False



""" Jubilee App class. """

import atexit, json, math, multiprocessing, os, platform, queue, signal, subprocess, sys, time
import __main__, numpy, pygame
from pygame import Rect
from pygame.font import Font
from pygame.surface import Surface
from pygame.mixer import Sound, Channel
from .worker import Worker
from .mode import Mode
from .base_classes import SpritePosition, Animation
from .misc import Config, Log, Color, Misc

class App:

	""" App class for app framework. """

	def __init__(self, workers=None):

		# single-instance check
		if self.check_running_process():
			print('Another instance of this script is already running.')
			sys.exit(0)

		# start log
		Log.backup()
		Log.info('Starting')

		# enable debug output if set
		if 'debug' in sys.argv:
			Log.set_file_level(Log.DEBUG)
		if 'console_debug' in sys.argv:
			Log.set_console_level(Log.DEBUG)

		# load config
		self.base_path = os.path.dirname(os.path.realpath(__main__.__file__))
		self.config_filename = os.path.join(self.base_path, 'config.txt')
		self.config = Config.load(self.config_filename, defaults=Worker.config_defaults)

		# init pygame
		self.headless = self.config.get('headless', False)
		if self.headless is True:
			Log.info('Running in headless mode')
			os.environ['SDL_VIDEODRIVER'] = 'dummy'
		self.nosound = self.config.get('nosound', False)
		if self.nosound is True:
			Log.info('Running in nosound mode')
			os.environ['SDL_AUDIODRIVER'] = 'dummy'
		try:
			pygame.init()
		except Exception as e:
			Log.error(f'Exception during pygame.init(): {e}')
			sys.exit(1)

		self.process_last = 0										# time of last process method
		self.process_period = 1.0 / max(1, int(self.config.get('app_process_fps', 10)))

		if self.headless is False:

			# initialize window
			self.screen_width, self.screen_height = self.config.get('screen_resolution', pygame.display.list_modes()[0])
			self.screen_center = int(self.screen_width / 2)
			self.screen_middle = int(self.screen_height / 2)
			if platform.uname().system == 'Darwin':
				pygame.mouse.set_visible(True)
				flags = pygame.DOUBLEBUF
			else:
				pygame.mouse.set_visible(False)
				flags = pygame.DOUBLEBUF | pygame.FULLSCREEN | pygame.NOFRAME
			size = (self.screen_width, self.screen_height)

			# adjust for screen rotation
			self.screen = None
			self.screen_rotation = self.config.get('screen_rotation', 0)
			if self.screen_rotation in (90, 270):
				size = (size[1], size[0])
			try:
				self.window = pygame.display.set_mode(size=size, flags=flags, display=0, vsync=1)
			except:
				self.window = pygame.display.set_mode(size=size, flags=flags, display=0)
			if self.screen_rotation != 0:
				self.screen = self.window
				self.window = pygame.Surface(size=(self.screen_width, self.screen_height))

			# fade overlay
			self.display_fade_overlay = Surface(size=size).convert_alpha()
			self.display_fade_steps = None
			self.display_fade_step = None
			self.display_fade_end_mode = None
			self.display_fade_end_parameters = None

			# drawing constants
			self.margin = 5													# margin between content and screen edge (px)
			self.button_margin = 3									# margin between buttons (px)
			self.button_border = 1									# width of button border (px)
			self.underscore_position = 7						# pixels under text for underscores
			self.popover_message = None							# current popover message
			self.popover_steps = 50									# steps
			self.popover_step = 0										# current popover step
			self.draw_last = 0											# time of last draw method
			self.fps_count = 0											# FPS count for last second
			self.fps_counting = 0										# FPS count for current second
			self.fps_time = int(time.time())				# time of current FPS count
			self.draw_period = 1.0 / max(1, int(self.config.get('app_draw_fps', 10)))

			# fonts
			self.standard_font_size = int(self.config.get('font_size', 11))
			# use a different default font size on MacOS
			if platform.uname().system == 'Darwin':
				self.standard_font_size = 15
			self.font_list = pygame.font.get_fonts()
			self.standard_font_name = None
			self.standard_font = None
			self.set_standard_font()
			self.display_fps = self.config.get('display_fps', False)

		# fetch initial events
		# this is necessary because if MacOS receives a mouse-click on the window before
		# calling events the first time, subsequent click events are often dropped.
		pygame.event.get()

		# add workers
		self.workers = {}
		if workers is not None:
			self.add_workers(workers)

		# modes
		self.modes = {}
		self.mode = None

		# app state
		self.app_state = {}
		self.app_state_filename = self.config.get('app_state_filename', 'app_state.txt')
		self.app_state_start_filename = self.config.get('app_state_start_filename', 'app_state_start.txt')
		self.persist_app_state = self.config.get('persist_app_state', True)
		if self.persist_app_state is True:
			self.load_app_state()

		# script information
		self.script = None

		# images
		self.images_path = os.path.join(self.base_path, 'images')
		self.images, self.animations = self.load_images(self.images_path)

		# sounds
		self.sounds_path = os.path.join(self.base_path, 'sounds')
		self.sounds = self.load_sounds(self.sounds_path)
		self.sound_volume = 100
		self.sound_retainer = None
		if self.config.get('sound_retainer', False) is True:
			self.set_sound_retainer(True)

		# music
		self.music_path = os.path.join(self.base_path, 'music')
		self.music_volume = 100
		self.music_fade_steps = None
		self.music_fade_step = None

		# pointer input
		self.pointer = None
		if self.headless is False:
			if platform.uname().system == 'Darwin':
				from .mouse_interface import MouseInterface
				self.pointer = MouseInterface()
			elif self.config.get('pointer_input', True) is True:
				from .touch_interface import TouchInterface
				scale = self.config.get('screen_scale', [[0, 319, -1], [0, 239, 1]])
				swap_axes = self.config.get('swap_axes', False)
				self.pointer = TouchInterface(resolution=(self.screen_width, self.screen_height), scale=scale, swap_axes=swap_axes)
			self.pointer_input_last = time.time()				# time of last pointer input
			self.pointer_input_debouncing = 100					# time between pointer inputs (ms)

		# key input
		self.new_keys = []														# all keys newly pressed this frame
		self.held_keys = []														# all keys currently pressed
		self.keyboard_buffering = False								# whether key input is being buffered
		self.keyboard_buffer = ''
		self.keyboard_buffer_chars = []

		# initialize script
		self.init_script()

		# app-specific initialization
		self.init()

	def init(self):
		""" App-specific initialization. """

	def run(self):
		""" Main run loop for application. """

		try:
			while True:
				if time.time() - self.process_last >= self.process_period:
					self.on_process()
				if self.headless is False:
					if time.time() - self.draw_last >= self.draw_period:
						self.on_draw()
				process_delay = self.process_period - (time.time() - self.process_last)
				draw_delay = 1 if self.headless is True else self.draw_period - (time.time() - self.draw_last)
				delay = min(process_delay, draw_delay)
				if delay > 0:
					time.sleep(delay)
		except Exception as e:
			Log.error(e)

	def add_worker(self, worker: type):
		""" Adds instance of worker class to app. """

		app_queue = multiprocessing.Queue()					# queue to send messages from app
		worker_queue = multiprocessing.Queue()			# queue to receive messages from worker
		config_manager = (len(self.workers) == 0)		# first worker manages config
		log_manager = (len(self.workers) == 0)			# first worker also manages log rotation
		if len(self.workers) > 0:
			time.sleep(1)															# delay to allow previous worker to start
		worker_instance = worker(app_queue, worker_queue, config_manager, log_manager)
		self.workers[worker_instance.name] = worker_instance

	def add_workers(self, workers: list):
		""" Adds workers. """

		for worker in workers:
			self.add_worker(worker)

	def add_mode(self, mode: type|Mode):
		""" Adds mode to mode list.

				Args:
					mode:								Mode name or subclass.
		"""

		if isinstance(mode, type):		# if mode is provided as a class, instantiate it
			mode = mode()
		mode.app = self
		mode.init()
		mode.load_resources()
		self.modes[mode.name] = mode
		if len(self.modes) == 1:
			self.set_mode(mode.name)

	def add_modes(self, modes: list):
		""" Adds modes. """

		for mode in modes:
			self.add_mode(mode)

	def set_mode(self, mode: str|Mode, mode_parameters: dict=None):
		""" Sets mode.

				Args:
					mode:								Mode name or object.
					mode_parameters:		Parameters for enter() for new mode.
		"""

		# insert previous_mode into mode_parameters
		mode_parameters = mode_parameters or {}
		mode_parameters['previous_mode'] = None if self.mode is None else self.mode.name

		# switch from mode
		if self.mode is not None:
			self.mode.on_exit()

		# switch to mode
		new_mode = self.modes.get(mode) if isinstance(mode, str) else mode
		if new_mode is None:
			Log.error(f'No known mode named {mode}')
			return
		self.mode = new_mode
		try:
			self.mode.on_enter(mode_parameters=mode_parameters)
		except Exception as e:
			Log.error(f'Exception while entering mode {self.mode.name}: {e}')
		submode = mode_parameters.get('submode')
		if submode is not None:
			self.mode.set_submode(submode)

	def on_process(self):
		""" Main app process event receiver. Calls current mode process method. """

		self.process()

		if self.mode is None:
			return

		try:
			self.process_last = time.time()
			self.handle_events()
			report_threshold = 0.2		# report processing if more than 0.2 seconds

			start = time.time()
			if self.config.get('modal') is True:			# only process current mdoe
				if self.mode is None:
					return
				try:
					self.mode.on_process()
				except Exception as e:
					Log.error(f'Error processing mode {self.mode.name}: {e}')
				duration = time.time() - start
				if duration > report_threshold:
					Log.info(f'Processing mode {self.mode.name} took {duration:.3f}')
			else:																			# process all modes
				mode_times = {}
				for mode in self.modes.values():
					mode_start = time.time()
					try:
						mode.on_process()
					except Exception as e:
						Log.error(f'Error processing mode {mode.name}: {e}')
					mode_times[mode.name] = f'{time.time() - mode_start:.3f}'
				duration = time.time() - start
				if duration > report_threshold:
					Log.info(f'Processing modes took {duration:.3f} s')
					Log.info(f'  Mode times: {mode_times}')
		except Exception as e:
			Log.error(e)

	def process(self):
		""" App process stub function. Called before mode-specific processing
				(in case the app needs to change modes). """

	def on_draw(self):
		""" Main draw method event receiver. Calls current mode draw method and
				then draws controls and popover message. """

		self.draw_last = time.time()

		try:

			self.fps_counting += 1
			current_time = int(time.time())
			if current_time != self.fps_time:
				self.fps_time = current_time
				self.fps_count = self.fps_counting
				self.fps_counting = 0

			# draw mode
			if self.mode is not None:
				try:
					self.mode.on_draw()
				except Exception as e:
					Log.error(f'Error drawing mode {self.mode.name}: {e}')

			# draw popover
			if self.popover_message is not None and self.popover_steps is not None:
				if self.popover_step < self.popover_steps:
					self.popover_step += 1
					width = min(int(self.screen_width * 0.8), 300)
					height = 100
					left = int(self.screen_center - width / 2)
					top = self.screen_middle - 50
					self.fill_rect(left, top, width, height, color='black')
					self.draw_rect(left, top, width, height, color='white')
					self.center_text(self.popover_message, color='white')
				else:
					self.popover_message = None
					self.popover_steps = None
					self.popover_step = None

			self.apply_display_fade()

			if self.display_fps:
				self.draw_text(str(self.fps_count), 20, 20, color='white', font=self.standard_font_sizes[18])

			# call app draw method
			self.draw()

			# flip display
			self.flip()

			# apply music fade
			self.apply_music_fade()

		except Exception as e:
			Log.error(e)

	def draw(self):
		""" Stub function for app draw method. """

	def handle_events(self):
		""" Handle pygame events and messages from worker. """

		events = list(pygame.event.get())
		if self.pointer:
			self.pointer.clicked = False
		for event in events:
			if event.type == pygame.QUIT:
				self.exit(0)
			elif event.type in (pygame.MOUSEBUTTONDOWN, pygame.MOUSEBUTTONUP) and self.pointer is not None:
				if self.pointer.handle_event(event):
					self.on_click(self.pointer.x, self.pointer.y)

		if self.pointer is not None:

			# detect events for touch pointer (via polling)
			if self.pointer.detect_events():
				# special function for click events to handle debouncing and mode input
				self.on_click(self.pointer.x, self.pointer.y)

			# respond to hold and release events, and then update state
			if self.pointer.down is True:
				if self.pointer.held is True:		# held
					self.mode.on_hold()
			elif self.pointer.held is True:		# released
				self.mode.on_release()

			# set state for next time
			self.pointer.held = self.pointer.down

		# register keyboard input
		if self.config.get('keyboard_input', True) is True:
			pressed_keys = pygame.key.get_pressed()
			held_keys = list(k for k in Misc.key_names if pressed_keys[pygame.key.key_code(k)])
			self.new_keys = list(k for k in held_keys if k not in self.held_keys)
			self.held_keys = held_keys

			# update keyboard buffer
			if self.keyboard_buffering is True:
				for k in self.new_keys:
					if k == 'backspace':
						if len(self.keyboard_buffer_chars) > 0:
							self.keyboard_buffer_chars = self.keyboard_buffer_chars[:-1]
							self.keyboard_buffer = self.keyboard_buffer[:-1]
					elif k not in ('return', 'left shift', 'left ctrl', 'right shift', 'right ctrl', 'f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7', 'f8', 'f9', 'f10', 'f11', 'f12', 'insert' ,'home', 'end', 'right', 'left', 'up', 'down', 'delete', 'escape'):
						self.keyboard_buffer_chars.append(k)
						if any(shift in self.held_keys for shift in ['left shift', 'right shift']):
							if k in Misc.key_shift_symbols:
								self.keyboard_buffer += Misc.key_shift_symbols[k]
						elif k in Misc.key_symbols:
							self.keyboard_buffer += Misc.key_symbols[k]

		self.receive_messages()

	# messaging with workers

	def send_message(self, message: str|dict, worker_name: str=None):
		""" Sends a message to worker, or first worker by default. """

		if len(self.workers) == 0:
			Log.error('No workers')
			return
		if worker_name is not None and worker_name not in self.workers:
			Log.error(f'No known worker {worker_name}')
			return
		if isinstance(message, str):
			message = {'action': message}
		first_worker = self.workers[list(self.workers.keys())[0]]
		worker = self.workers.get(worker_name, first_worker)
		try:
			worker.app_queue.put(json.dumps(message))
		except Exception as e:
			Log.error(e)

	def receive_messages(self):
		""" Receives messages from workers and calls process_message() on each. """

		for worker in self.workers.values():
			try:
				while True:
					message = worker.worker_queue.get_nowait()
					self.process_message(json.loads(message), sender=worker.name)
			except queue.Empty:
				continue
			except Exception as e:
				Log.error(e)
				continue

	def process_message(self, message: dict, sender: str=None):
		""" Processes a message from worker. """

		action = message.get('action')
		if action == 'exit':
			self.exit(0)
		elif action == 'config updated':
			self.config = message.get('config', {})
			for name in (name for name, worker in self.workers.items() if worker.config_manager is False):
				self.send_message(message, name)
			if self.headless is False:
				self.set_standard_font()
		else:
			Log.error(f'Received unknown message: {message}')

	# low-level graphics methods

	def flip(self):
		""" Flips display buffers. This occurs automatically at the end of draw().
				Also handles software rotation. """

		if self.screen_rotation != 0:
			rotated_screen = pygame.transform.rotate(self.window, self.screen_rotation)
			self.screen.blit(rotated_screen, (0, 0))
		pygame.display.flip()

	def fill_screen(self, color: str|int|tuple='white'):
		""" Fills screen with indicated color. """

		self.fill_color(color, dest=self.window)

	def fill_color(self, color: str|int|tuple='white', dest: Surface=None):
		""" Fills surface with indicated color. """

		dest = dest or self.window
		color = Misc.get_color(color)
		size = dest.get_size()
		dest.fill(color, (0, 0, size[0], size[1]))

	def fill_static(self, dest: Surface=None):
		""" Fills surface with white-noise static. """

		dest = dest or self.window
		pygame.surfarray.blit_array(dest, numpy.random.randint(255, size=dest.get_size()) * 65793)

	def draw_text(self, text: str, x: int|float, y: int|float, color='white', font: str|Font=None, alignment: str='left', dest: Surface=None):
		""" Draws text. Font can be either a pygame.font.SysFont, a font name (default size),
				or None for default font. """

		dest = dest or self.window
		if isinstance(font, Font) is False:
			font = pygame.font.SysFont(font, self.standard_font_size) if font in self.font_list else self.standard_font
		color = Misc.get_color(color)
		surface = font.render(text, True, color)
		(width, height) = surface.get_size()
		top = int(y - height / 2)
		left = int(x)
		if alignment == 'center':
			left = int(x - width / 2)
		elif alignment == 'right':
			left = int(x - width)
		dest.blit(surface, (left, top))

	def center_text(self, text: str, y: int|float=None, color='white', font: str|Font=None, dest: Surface=None):
		""" Draws horizontally (and, optionally, vertically) centered text. """

		dest = dest or self.window
		x = dest.get_size()[0] / 2
		y = dest.get_size()[1] / 2 if y is None else y
		self.draw_text(text, x, y, color=color, font=font, alignment='center', dest=dest)

	def get_text_size(self, text: str, font: str|Font=None):
		""" Gets size of rendered text. Font can be either a pygame.font.SysFont, a font name
				(default size), or None for default font. """

		if isinstance(font, Font) is False:
			font = pygame.font.SysFont(font, self.standard_font_size) if font in self.font_list else self.standard_font
		surface = font.render(text, True, Color.WHITE.value)
		(width, height) = surface.get_size()
		return (width, height)

	def draw_pixel(self, x: int|float, y: int|float, color='white', dest: Surface=None):
		""" Draw pixel at coordinate. """

		dest = dest or self.window
		color = Misc.get_color(color)
		dest.set_at((int(x), int(y)), color)

	def draw_line(self, x1: int|float, y1: int|float, x2: int|float, y2: int|float, width: int|float=1, color='white', dest: Surface=None):
		""" Draws pixel at coordinate. """

		dest = dest or self.window
		color = Misc.get_color(color)
		pygame.draw.line(dest, color, (x1, y1), (x2, y2), width)

	def draw_rect(self, left: int|float, top: int|float, width: int|float, height: int|float, line_width: int|float=1, color='white', dest: Surface=None):
		""" Draws rect at coordinates. """

		dest = dest or self.window
		color = Misc.get_color(color)
		pygame.draw.rect(dest, color, (left, top, width, height), line_width)

	def fill_rect(self, left: int|float, top: int|float, width: int|float, height: int|float, color='white', dest: Surface=None):
		""" Fills pixel at coordinates. """

		dest = dest or self.window
		color = Misc.get_color(color)
		dest.fill(color, (left, top, width, height))

	def draw_polygon(self, coordinates: list, width: int|float=1, color='white', dest: Surface=None):
		""" Draws polygon at coordinates. """

		dest = dest or self.window
		color = Misc.get_color(color)
		pygame.draw.polygon(dest, color, coordinates, width=width)

	def draw_circle(self, x: int|float, y: int|float, radius: int|float=1, width: int|float=1, color='white', dest: Surface=None):
		""" Draws circle centered at coordinate. """

		dest = dest or self.window
		color = Misc.get_color(color)
		pygame.draw.circle(dest, color, (x, y), radius, width=width)

	def fill_circle(self, x: int|float, y: int|float, radius: int|float=1, color='white', dest: Surface=None):
		""" Fills circle centered at coordinate. """

		self.draw_circle(x, y, radius=radius, width=0, color=color, dest=dest)

	def draw_arc(self, x: int|float, y: int|float, width: int|float, height: int|float, start_angle: int|float, stop_angle: int|float, line_width: int|float=1, color='white', dest: Surface=None):
		""" Draws arc at x/y coordinate from start_angle (degrees) to stop_angle (degrees). """

		dest = dest or self.window
		color = Misc.get_color(color)
		pygame.draw.arc(dest, color, Rect(x, y, width, height), start_angle * math.pi / 180, stop_angle * math.pi / 180, width = line_width)

	# surface and image methods

	def create_surface(self, x: int|float, y: int|float, color='black', alpha_blend: bool=False, flags: int=None) -> Surface:
		""" Creates a new pygame surface with given dimensions.
				Alpha blend specifies a surface with alpha transparency. """

		x = int(x); y = int(y); flags=flags or 0
		if alpha_blend:
			flags = flags | pygame.SRCALPHA
		surface = Surface((x, y), flags=flags)
		surface = surface.convert_alpha() if alpha_blend else surface.convert()
		if color is not None:
			c = Misc.get_color(color)
			surface.fill(c, (0, 0, x, y))
		return surface

	def copy_surface(self, surface: Surface):
		""" Returns a copy of the surface. """

		return surface.copy()

	def load_images(self, path: str) -> (dict, dict):
		""" Loads images and animations from path. Called here and also in Mode. """

		images = {}; animations = {}
		image_types = ['.jpg', '.jpeg', '.png', '.bmp']
		if os.path.isdir(path) is False:
			return images, animations

		# load images
		for filename in (f for f in os.listdir(path) if os.path.splitext(f)[1].lower() in image_types):
			full_filename = os.path.join(path, filename)
			name = os.path.splitext(filename)[0].lower()
			try:
				image = self.load_image(full_filename)
				if image is not None:
					images[name] = image
				else:
					Log.error(f'Image {full_filename} failed to load')
			except Exception as e:
				Log.error(f'Exception loading image {full_filename}: {e}')

		# load animations
		for animation_name in list(d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))):
			animation = Animation(name=animation_name)
			animation_names = []
			files = sorted(f for f in os.listdir(os.path.join(path, animation_name)) if os.path.splitext(f)[1].lower() in image_types)
			for i, filename in enumerate(files):
				image = self.load_image(os.path.join(path, animation_name, filename))
				if image is not None:
					animation.frames.append(image)
					name = os.path.splitext(filename)[0].lower()
					animation_names.append(name)
			if len(animation.frames) == 0:
				continue
			# define sequences
			sequence_files = {}
			for i, name in enumerate(animation_names):
				if '_' not in name:
					animation.sequences[name] = [i]
				else:		# parse _ for sequence number
					sequence_name, sequence_number = name.rsplit('_', 1)
					sequence_files.setdefault(sequence_name, {})
					try:
						sequence_files[sequence_name][int(sequence_number)] = i
					except:		# file ends with non-numeric field
						animation.sequences[name] = [i]

			# once all sequences are loaded, sort by names and store in animation.sequences
			for sequence_name, files in sequence_files.items():
				sequence_numbers = sorted(files.keys())
				animation.sequences[sequence_name] = list(files[s] for s in sequence_numbers)

			# add animation to animations library
			animations[animation_name.lower()] = animation

		return images, animations

	def load_image(self, filename: str, alpha_blend: bool=False) -> Surface|None:
		""" Loads image at filename. Specify Alpha Blend if the image has an alpha channel. """

		if os.path.isfile(filename) is False:
			Log.error(f'No image named {filename}')
			return None
		try:
			alpha_blend = alpha_blend or (os.path.splitext(filename)[1].lower() == '.png')
			image = pygame.image.load(filename)
			image = image.convert_alpha() if alpha_blend else image.convert()
			return image
		except Exception as e:
			Log.error(e)
			return None

	def get_image(self, image: str|Surface) -> Surface|None:
		""" Finds an image by name or path. """

		if image is None or isinstance(image, Surface):
			return image
		if isinstance(image, str) is False:
			Log.error(f'Could not identify image of type {type(image)}')
			return None
		# in order: check mode image library, then app image library, and lastly try to load
		i = self.mode.images.get(image) if self.mode is not None else None
		i = i or self.images.get(image)
		i = i or self.load_image(image)
		return i

	def get_animation(self, animation: str|Animation) -> Animation|None:
		""" Finds an animation by name or path. """

		if animation is None or isinstance(animation, Animation):
			return animation
		if isinstance(animation, str) is False:
			Log.error(f'Could not identify nimation of type {type(animation)}')
			return None
		# in order: check mode image library, then app image library
		a = self.mode.animations.get(animation) if self.mode is not None else None
		a = a or self.animations.get(animation)
		return a

	def blit(self, image: str|Surface, x: int|float, y: int|float, position: SpritePosition=None, scale: int|float|tuple=None, area: Rect=None, flags: int=None, dest: Surface=None):
		""" Blits image to dest at coordinates.
				Optionally scale by factor or to (width, height).
				Optionally specify an area of the source image (post-scaling).
				Surfaces that have an alpha channel are blitted with pygame.BLEND_ALPHA_SDL2. """

		i = self.get_image(image)
		if i is None:
			Log.error(f'Could not get image of type {image}')
			return
		flags = flags or 0
		dest = dest or self.window
		if i.get_alpha() is not None:
			flags = flags | pygame.BLEND_ALPHA_SDL2
		try:
			if scale is not None:
				if isinstance(scale, tuple):
					i = self.scale_image(i, scale[0], scale[1])
				else:
					i = self.scale_image(i, scale)
			if position == SpritePosition.Center:
				size = i.get_size()
				x -= size[0] / 2; y -= size[1] / 2
			elif position == SpritePosition.Bottom:
				size = i.get_size()
				x -= size[0] / 2; y -= size[1]
			dest.blit(i, (x, y), area=area, special_flags=flags)
		except Exception as e:
			Log.error(e)

	def scale_image(self, image: str|Surface, x_scale: int|float, y_scale: int|float=None) -> Surface|None:
		""" Scales image. Scale proportionally (omit y_scale) or disproportionately. """

		i = self.get_image(image)
		if i is None:
			Log.error(f'Could not get image of type {image}')
			return None
		x, y = image.get_size()
		x = x * x_scale
		y = y * (y_scale or x_scale)
		return pygame.transform.smoothscale(i, (x, y))

	def flip_image(self, image: str|Surface, horizontal: bool=False, vertical: bool=False) -> Surface|None:
		""" Flips image. """

		i = self.get_image(image)
		if i is None:
			Log.error(f'Could not get image of type {image}')
			return None
		return pygame.transform.flip(i, horizontal, vertical)

	def rotate_image(self, image: str|Surface, degrees: int|float) -> Surface|None:
		""" Rotates image. """

		i = self.get_image(image)
		if i is None:
			Log.error(f'Could not get image of type {image}')
			return None
		return pygame.transform.rotate(i, degrees)

	def shift_image_hue(self, image: str|Surface, delta: int|float) -> Surface|None:
		""" Shifts image hue by a delta value (range 0-360). """

		i = self.get_image(image)
		if i is None:
			Log.error(f'Could not get image of type {image}')
			return None
		i = i.copy()			# make a copy of the image to edit pixels in place
		pixels = pygame.PixelArray(i)
		for x in range(i.get_width()):
			for y in range(i.get_height()):
				color = i.unmap_rgb(pixels[x][y])
				h, s, l, a = color.hsla
				color.hsla = (int(h) + int(delta)) % 360, int(s), int(l), int(a)
				pixels[x][y] = color
		del pixels
		return i

	def set_popover(self, message: str, steps: int=None):
		""" Creates a popover message. """

		self.popover_message = message
		self.popover_steps = steps if steps is not None else int(self.config.get('app_draw_fps', 10) * 0.75)
		self.popover_step = 0

	def start_display_fade(self, steps: int=None, color='black', end_mode: str=None, end_parameters: dict=None):
		""" Sets fade-in or fade-out.
				Positive steps = fade in from solid and increase over specified steps.
				Negative steps = fade out to solid over specified steps.
				Optionally, call set_mode with the specified mode and parameters at end of fade.
				None or zero steps will cancel current fade.
		"""

		if steps is None or steps == 0:
			self.display_fade_steps = None
			self.display_fade_step = None
			self.display_fade_end_mode = None
			self.display_fade_end_parameters = None
			return
		c = Misc.get_color(color)
		self.display_fade_overlay.fill(c, (0, 0, self.screen_width, self.screen_height))
		self.display_fade_steps = steps
		self.display_fade_step = 0
		self.display_fade_end_mode = end_mode
		self.display_fade_end_parameters = end_parameters

	def apply_display_fade(self):
		if self.display_fade_steps is None or self.display_fade_step is None:
			return
		self.display_fade_step += 1
		total_steps = abs(self.display_fade_steps)
		fade = min(1.0, self.display_fade_step / total_steps)
		alpha = int(255 * (fade if self.display_fade_steps < 0 else (1.0 - fade)))
		self.display_fade_overlay.set_alpha(alpha)
		self.window.blit(self.display_fade_overlay, (0, 0), special_flags=pygame.BLEND_ALPHA_SDL2)
		if self.display_fade_step >= total_steps:
			end_mode = self.display_fade_end_mode
			end_parameters = self.display_fade_end_parameters
			self.display_fade_steps = None
			self.display_fade_step = None
			self.display_fade_end_mode = None
			self.display_fade_end_parameters = None
			if end_mode is not None:
				self.set_mode(end_mode, end_parameters)

	# sound functions

	def load_sounds(self, path: str) -> dict:
		""" Loads sounds from path. Called here and also in Mode. """

		sounds = {}
		sound_types = ['.wav']
		if os.path.isdir(path) is False:
			return sounds
		for filename in (f for f in os.listdir(path) if os.path.splitext(f)[1].lower() in sound_types):
			full_filename = os.path.join(path, filename)
			name = os.path.splitext(filename)[0].lower()
			try:
				sound = self.load_sound(full_filename)
				if sound is not None:
					sounds[name] = sound
				else:
					Log.error(f'Sound {full_filename} failed to load')
			except Exception as e:
				Log.error(f'Exception loading sound {full_filename}: {e}')
		return sounds

	def load_sound(self, sound) -> Sound:
		""" Loads a sound by filename or buffer and returns pygame Sound object. """

		if isinstance(sound, str) and not os.path.isfile(sound):
			Log.error(f'No file named {sound}')
			return None
		try:
			return pygame.mixer.Sound(sound)
		except Exception as e:
			Log.error(e)
			return None

	def get_sound(self, sound: str|Sound):
		""" Finds a sound by name. """

		if sound is None or isinstance(sound, Sound):
			return sound
		if isinstance(sound, str) is False:
			Log.error(f'Could not identify sound of type {type(sound)}')
			return None
		# in order: check mode sound library, then app sound library, and lastly try to load
		s = self.mode.sounds.get(sound, None) if self.mode is not None else None
		s = s or self.sounds.get(sound, None)
		s = s or self.load_sound(sound)
		return s

	def play_sound(self, sound: str|Sound, loops: int=None, volume: int|float=None):
		""" Plays sound. Sound can be a pygame Sound, the filename of a sound
				in the mode sound library or app sound library, or a filename.
				Volume can be specified (range 0.0-1.0) to override
				sound volume; otherwise, sound will play at preset sound volume. """

		if self.nosound is True:
			return
		sound = self.get_sound(sound)
		if sound is None:
			Log.error(f'Could not get sound {sound}')
			return
		volume = volume or self.sound_volume
		if volume is not None and volume != 100:
			sound.set_volume(volume / 100.0)
		sound.play(loops=loops or 0)

	def play_sound_on_channel(self, sound: str|Sound, loops: int=None, volume: int|float=None) -> Channel|None:
		""" Plays sound on open channel. Same logic as play_sound(),
				except that it returns the channel of the sound. """

		if self.nosound is True:
			return None
		sound = self.get_sound(sound)
		if sound is None:
			Log.error(f'Could not get sound {sound}')
			return None
		volume = volume or self.sound_volume
		if volume is not None and volume != 100:
			sound.set_volume(volume / 100.0)
		channel = pygame.mixer.find_channel()
		if channel is None:
			Log.error('Could not find open channel')
			return None
		channel.play(sound, loops=loops or 0)
		return channel

	def set_volume(self, volume: int|float, sound_volume: int|float=None):
		""" Set volume level (range 0-100). Specifying one value will
				change volume level for both music and sound; specifying
				two values will set separate volume for music and sound.
				Volume changes for currently playing music will occur
				immediately, but currently playing sounds are not affected. """

		if self.nosound is True:
			return

		self.music_volume = volume
		pygame.mixer.music.set_volume(volume / 100.0)
		self.sound_volume = sound_volume or volume

	def set_sound_retainer(self, enable: bool=True):
		""" Sets a retainer sound that plays a 1-second loop of very
				quiet noise to keep a sound channel active. """

		if self.nosound is True:
			return

		if enable:
			filename = os.path.join(self.base_path, 'sounds', 'sound_retainer.wav')
			if os.path.isfile(filename) is False:
				Log.error(f'{filename} does not exist')
				return
			self.sound_retainer = pygame.mixer.Sound(filename)
			self.sound_retainer.play(loops=-1)
		elif self.sound_retainer is not None:
			self.sound_retainer.stop()
			self.sound_retainer = None

	# music functions

	def get_music(self, music_name: str):
		""" Finds music by name. """

		if music_name is None or isinstance(music_name, str) is False:
			Log.error(f'Could not identify music of type {type(music_name)}')
			return None

		mode_path = os.path.join(self.base_path, self.mode.name, 'music', music_name)
		app_path = os.path.join(self.base_path, 'music', music_name)
		music = next(filter(os.path.isfile, (mode_path, app_path, music_name)), None)
		return music

	def play_music(self, filename: str, loops: int=0, volume: int|float=None):
		""" Plays music file.
				Loops can indicate a specific number, or -1 for indefinite repeat.
				Volume can be specified for this music (range 0-100) or will play at default.
				"""

		if self.nosound is True:
			return
		music = self.get_music(filename)
		if music is None:
			Log.error(f'No file named {filename}')
			return
		volume = volume or self.music_volume
		pygame.mixer.music.set_volume(volume / 100.0)
		try:
			pygame.mixer.music.load(music)
			pygame.mixer.music.play(loops=loops)
		except Exception as e:
			Log.error(e)

	def stop_music(self):
		""" Stops music. """

		if self.nosound is True:
			return

		pygame.mixer.music.stop()
		self.music_fade_steps = self.music_fade_step = None

	def is_music_playing(self):
		""" Returns whether music is playing. """

		if self.nosound is True:
			return False
		return pygame.mixer.music.get_busy()

	def start_music_fade(self, steps: int):
		""" Starts a music fade over a given number of steps. """

		self.music_fade_steps = steps
		self.music_fade_step = 0

	def apply_music_fade(self):
		""" Applies music fade, if it exists. """

		if self.music_fade_steps is None or self.music_fade_step is None:
			return

		if pygame.mixer.music.get_busy() is False:
			self.music_fade_steps = self.music_fade_step = None
			return

		self.music_fade_step += 1
		level = max(0.0, 1.0 - (self.music_fade_step / float(self.music_fade_steps)))
		pygame.mixer.music.set_volume((self.music_volume * level) / 100.0)
		if self.music_fade_step >= self.music_fade_steps:
			pygame.mixer.music.stop()
			pygame.mixer.music.set_volume(self.music_volume / 100.0)
			self.music_fade_steps = self.music_fade_step = None

	# app state methods

	def load_app_state(self):
		""" Loads app state. """

		filename = self.app_state_filename
		if os.path.isfile(filename) is False:
			filename = self.app_state_start_filename
			if os.path.isfile(filename) is False:
				Log.info('No app_state or app_state_start - using empty state dict')
				self.app_state = {}
				return
		with open(filename, 'rt', encoding='UTF-8') as f:
			try:
				self.app_state = json.loads(f.read().strip())
				Log.info('Loaded app state')
			except Exception as e:
				Log.error(e)
				self.app_state = {}

	def set_app_state(self, key, value):
		""" Sets app state key/value pair and saves state. """

		self.app_state[key] = value
		if self.persist_app_state is True:
			self.save_app_state()

	def save_app_state(self):
		""" Saves script state. """

		with open(self.app_state_filename, 'wt', encoding='UTF-8') as f:
			f.write(json.dumps(self.app_state, ensure_ascii=True))
		Log.debug('Saved app state')

	# script and scene methods

	def init_script(self):
		""" Loads script and script state. """

		# load script
		self.script = []
		script_filename = os.path.join(self.base_path, 'script.txt')
		if os.path.isfile(script_filename) is False:
			return
		with open(script_filename, 'rt', encoding='UTF-8') as f:
			lines = list(line.strip() for line in f.readlines() if len(line.strip()) > 0 and line[0] != '#')
			for line in lines:
				tokens = list(t.strip() for t in line.split() if len(t.strip()) > 0)
				token_pairs = list(t.split('=') for t in tokens if len(t.split('=')) == 2)
				if len(token_pairs) > 0:
					token_pairs = {t[0].strip(): t[1].strip() for t in token_pairs}
					self.script.append(token_pairs)

		# set scene number in app state
		self.app_state.setdefault('scene', 0)

		Log.debug('Loaded and initialized script')

	def run_script(self):
		""" Begins script execution. """

		self.select_scene(self.app_state['scene'])

	def select_scene(self, scene_id: str|int):
		""" Selects scene by name or number. """

		if self.script is None:
			Log.error('No loaded script')
			return
		scene_number = None
		if isinstance(scene_id, int):
			if scene_id >= len(self.script):
				Log.error(f'Scene number {scene_id} cannot be selected ({len(self.script)} scenes)')
				return
			scene_number = scene_id
		else:
			matches = list(i for i, s in enumerate(self.script) if s.get('name') == scene_id)
			if len(matches) == 0:
				Log.error(f'Unknown scene {scene_id}')
				return
			scene_number = matches[0]
		self.set_app_state('scene', scene_number)
		scene = self.script[scene_number]
		mode_name = scene.get('mode')
		if mode_name not in self.modes:
			Log.error(f'No mode named {mode_name}')
			return
		self.set_mode(mode_name, mode_parameters=scene.copy())
		Log.debug(f'Selected scene {scene_number} ({scene})')

	def advance_scene(self, delta: int=1):
		""" Advances to indicated scene. """

		self.select_scene(self.app_state['scene'] + delta)

	# pointer input methods

	def on_click(self, x: int|float, y: int|float):
		""" Mode click event handler. """

		# debouncing check
		if self.pointer_input_last is not None and time.time() < self.pointer_input_last + self.pointer_input_debouncing / 1000:
			return

		if x is None or y is None:
			return

		self.pointer_input_last = time.time()
		self.mode.on_click(x, y)

	# keyboard input methods

	def start_keyboard_buffering(self):
		""" Starts keyboard buffering and clears buffer. """

		self.keyboard_buffering = True
		self.clear_keyboard_buffer()

	def clear_keyboard_buffer(self):
		""" Resets keyboard buffer for new text input. """

		self.keyboard_buffer_chars = []
		self.keyboard_buffer = ''

	def stop_keyboard_buffering(self):
		""" Stops keyboard buffering. """

		self.keyboard_buffering = False

	# other primitives

	def create_rect(self, width: int|float, height: int|float):
		""" Returns a pygame.Rect. """

		return Rect(0, 0, int(width), int(height))

	def create_font(self, name='Arial', size=12):
		""" Creates a font from a path or a system font name. """

		if os.path.isfile(name):
			try:
				return pygame.font.SysFont(name, size)
			except Exception as e:
				Log.error(f'Exception loading font {name}: {e}')
				return None
		else:
			try:
				return pygame.font.SysFont(name, size)
			except Exception as e:
				Log.error(f'Exception creating font {name}: {e}')
				return None

	def set_standard_font(self):
		""" Sets default font, including a range of sizes. """

		self.standard_font_name = self.config.get('font', None) or self.font_list[0]
		self.standard_font = pygame.font.SysFont(self.standard_font_name, self.standard_font_size)
		self.standard_font_sizes = {i: pygame.font.SysFont(self.standard_font_name, i) for i in range(4, 65)}

	def change_font(self):
		""" Chooses next font in the font list as the default font. """

		current_index = len(self.font_list) if self.standard_font_name not in self.font_list else self.font_list.index(self.standard_font_name)
		new_standard_font = self.font_list[(current_index + 1) % len(self.font_list)]
		self.update_config('font', new_standard_font)
		Log.info(f'Changed font to {new_standard_font}')

	def update_config(self, key, value):
		""" Updates config key/value pair by sending a message to worker. """

		message = {'action': 'update config', 'key': key, 'value': value}
		for name in (name for name, worker in self.workers.items() if worker.config_manager):
			self.send_message(message, name)
		Log.info(f'Setting {key} to {value}')

	# exit functions

	def register_exit_handlers(self):
		""" Registers handlers to ensure that pygame.quit() is called. """

		atexit.register(self.exit)
		for s in [signal.SIGABRT, signal.SIGINT, signal.SIGTERM]:
			signal.signal(s, lambda *_: sys.exit(0))

	def exit(self, code=0):
		""" Exits application. """

		Log.info('Exiting')
		for worker in self.workers.values():
			if worker.worker_process is not None:
				self.send_message('exit', worker.name)
		if self.pointer is not None:
			self.pointer.release()
		try:
			pygame.quit()
		finally:
			sys.exit(code)

	def reboot(self):
		""" Reboots device. """

		Log.info('Rebooting device')
		self.fill_screen()
		self.flip()
		os.system('sudo shutdown -r now')

	def shut_down(self):
		""" Shuts down device. """

		Log.info('Shutting down device')
		self.fill_screen()
		self.flip()
		os.system('sudo shutdown now')

	@staticmethod
	def check_running_process(process_name: str=None) -> bool:
		""" Checks for another running process of the same name. """

		script_name = os.path.basename(__main__.__file__)
		with subprocess.Popen(f'ps -ef | grep -v grep | grep {process_name or script_name}', shell=True, stdout=subprocess.PIPE) as ps:
			output = ps.stdout.read(); ps.stdout.close(); ps.wait()
			output = list(o.strip() for o in output.decode('UTF-8').split('\n') if len(o.strip()) > 0)
			running_processes = list(list(o.strip() for o in line.split(' ') if len(o.strip()) > 0) for line in output if 'python' in line)
		return len(running_processes) > 1

if __name__ == '__main__':
	App().run()



""" Jubilee touch interface class using evdev. """

import evdev
from .base_classes import PointerInterface
from .misc import Log

class TouchInterface(PointerInterface):
	""" Touch interface class. """

	def __init__(self, resolution: list=None, scale: list=None, swap_axes: bool=False):
		super().__init__()
		try:
			# probe /dev/input/event* to determine which one has bustype 24
			device_number = None
			for i in range(0, 10):
				try:
					d = evdev.InputDevice(f'/dev/input/event{i}')
					if d.info.bustype == 24:
						device_number = i
						break
				except:
					pass
			if device_number is None:
				Log.error('Could not find touchscreen input among device events')
				self.touch = None
				return
			Log.info(f'Found touchscreen on device /dev/input/event{device_number}')
			device = f'/dev/input/event{device_number}'
			self.touch = evdev.InputDevice(device)
			self.touch.grab()
			Log.info(f'Grabbed {device} - info: {self.touch.info}')
		except Exception as e:
			Log.error(f'Exception during grab: {e}')
			return
		self.resolution = resolution
		self.scale = scale
		if resolution is None or scale is None:
			Log.warning('resolution and/or scale not specified')
		self.swap_axes = swap_axes

	def detect_events(self) -> bool:
		""" Detect touch events. """

		touched = False
		try:
			for event in self.touch.read():
				if event.type == evdev.ecodes.EV_ABS:
					if event.code == 54:
						self.x = int(((event.value - self.scale[0][0]) / (self.scale[0][1] - self.scale[0][0]) * self.scale[0][2] - (self.scale[0][2] - 1) / 2) * self.resolution[0])
					elif event.code == 53:
						self.y = int(((event.value - self.scale[1][0]) / (self.scale[1][1] - self.scale[1][0]) * self.scale[1][2] - (self.scale[1][2] - 1) / 2) * self.resolution[1])
				elif event.type == evdev.ecodes.EV_KEY and event.code == 330 and event.value == 1:
					if self.x is not None and self.y is not None:

						if self.swap_axes is True:
							x = self.y
							y = self.x
							self.y = y
							self.x = x

						self.down = True
						touched = True
				elif event.type == evdev.ecodes.EV_KEY and event.code == 330 and event.value == 0:
					self.down = False
					self.x = None
					self.y = None
		except:
			pass
		return touched

	def release(self):
		""" Release touch interface. """

		self.touch.ungrab()



""" Jubilee controls. """

from pygame.font import Font
from .misc import Log

class Control:
	""" Jubilee user control base class. """

	def __init__(self, x: int|float, y: int|float, width: int|float, height: int|float,
			click=None, hold=None, release=None, name: str=None, parameters: dict=None):
		self.app = None			# this is set in mode.add_control()
		self.x = x
		self.y = y
		self.width = width
		self.height = height
		self.provided_click_handler = click
		self.provided_hold_handler = hold
		self.provided_release_handler = release
		self.name = name or 'Control'
		self.visible = True
		self.enabled = True
		self.parameters = parameters or {}

	def bind(self, app):
		""" Control bind method for adding to a mode. """

		self.app = app

	def collide(self, x: int|float, y: int|float):
		""" Control collision detection method. """

		return (self.x <= x < self.x + self.width and self.y <= y < self.y + self.height)

	def on_click(self):
		""" Control click event receiver. """

		if self.provided_click_handler is not None:
			self.provided_click_handler()
	
	def click(self):
		""" Control click event stub method. """

	def on_hold(self):
		""" Control hold event receiver. """

		if self.provided_hold_handler is not None:
			self.provided_hold_handler()

	def on_release(self):
		""" Control release event receiver. """

		if self.provided_release_handler is not None:
			self.provided_release_handler()

	def draw(self):
		""" Control draw stub method. """

	def exit_handler(self):
		""" Control handler method for exiting app. """

		self.app.exit()

class LabeledControl(Control):
	""" Wrapper to add a left-side label to a control.
			Optionally specify an x offset for the control; default uses text width.
	"""

	def __init__(self, caption: str, control: Control, offset: int=None,
		font: Font|str=None, color='white', name: str=None, parameters: dict=None):
		super().__init__(control.x, control.y, control.width, control.height,
			name=name, parameters=parameters)
		self.control = control
		self.caption = caption
		self.label_font = font
		self.color = color
		self.offset = offset
		self.name = name or f'LabeledControl: {self.control.name}'
		self.is_initilized = False

	def bind(self, app):
		""" Control bind method for adding to a mode. """

		super().bind(app)
		self.set_layout()
		if self.control is not None:
			self.control.bind(app)

	def set_layout(self):
		""" Sets position of control based on layout. """

		offset = self.offset
		label_height = self.control.height
		if offset is None:		# calculate offset from caption size
			try:
				offset, label_height = self.app.get_text_size(self.caption, font=self.label_font)
			except Exception as e:
				Log.error(f'get_text_size failed: {e}')
				return
		self.control.x = self.x + offset
		self.width = offset + self.control.width
		self.height = max(label_height, self.control.height)

	def set_caption(self, caption: str):
		""" Sets caption. """

		self.caption = caption
		self.set_layout()

	def collide(self, x: int|float, y: int|float):
		""" LabeledControl collide function - tests only against underlying control. """

		return self.control.collide(x, y)

	def on_click(self):
		""" LabeledControl on_click event receiver. """

		self.control.on_click()

	def on_hold(self):
		""" LabeledControl on_hold event receiver. """

		self.control.on_hold()

	def on_release(self):
		""" LabeledControl on_release event receiver. """

		self.control.on_release()

	def draw(self):
		""" LabeledControl draw function. """

		try:
			y = self.control.y + (self.control.height / 2)
			self.app.draw_text(self.caption, self.x, y, color=self.color, font=self.label_font, alignment='left')
			self.control.draw()
		except Exception as e:
			Log.error(e)

class Button(Control):
	""" Button user control. """

	def __init__(self, caption: str, x: int|float, y: int|float, width: int|float,
			height: int|float, target_mode=None, target_mode_parameters: dict=None,
			click=None, hold=None, release=None, app_exit: bool=False,
			font: Font|str=None, color='white', background_color='black',
			name: str=None, parameters: dict=None):
		super().__init__(x, y, width, height, click=click, hold=hold, release=release,
			name=name, parameters=parameters)
		# note: the provided_click_handler is saved so that if user later sets or changes
		# target_mode or app_exit, the button-specific click handler included in this class
		# will correctly hand off to the correct function.
		# specifying target_mode or app_exit will override click
		self.caption = caption
		self.target_mode = target_mode
		self.target_mode_parameters = target_mode_parameters
		self.app_exit = app_exit
		self.font = font
		self.color = color
		self.name = name or f'Button: {caption}'
		self.background_color = background_color

	def on_click(self):
		""" Button click event receiver. """

		if self.target_mode:
			self.app.set_mode(self.target_mode, mode_parameters=self.target_mode_parameters)
		elif self.app_exit:
			self.app.exit()
		else:
			super().on_click()

	def draw(self):
		""" Button draw method. """

		try:
			if self.background_color is not None:
				self.app.fill_rect(self.x, self.y, self.width, self.height, color=self.background_color)
			self.app.draw_rect(self.x, self.y, self.width, self.height, line_width=self.app.button_border, color=self.color)
			x = int(self.x + self.width / 2)
			y = int(self.y + self.height / 2)
			self.app.draw_text(self.caption, x, y, color = self.color, font = self.font, alignment = 'center')
		except Exception as e:
			Log.error(e)

class HoldButton(Button):
	""" Hold button user control that calls .click() after a hold period.
			This is a wrapper class for the basic button. This class accepts the usual button
			parameters, including target_mode and app_exit, but overrides click and only
			calls it when .hold_steps == .hold_step. A completion of the hold event calls
			the Button click function, which handles target_mode and app_exit.
			hold_steps indicates how many frames the hold should take for activation.
			hold_color indicates the color to fill the button as a progress bar. """

	def __init__(self, caption: str, x: int|float, y: int|float, width: int|float,
			height: int|float, hold_color='red', hold_steps: int=12, click=None, hold=None,
			release=None, target_mode=None, target_mode_parameters: dict=None,
			app_exit: bool=False, font: Font|str=None, color='white', background_color=None,
			name: str=None, parameters: dict=None):

		super().__init__(caption, x, y, width, height, target_mode=target_mode,
			target_mode_parameters=target_mode_parameters,
			click=click, hold=hold, release=release,
			app_exit=app_exit, font=font, color=color, background_color=background_color,
			name=name, parameters=parameters)
		self.hold_steps = max(1, int(hold_steps))
		self.hold_step = 0
		self.hold_color = hold_color
		self.name = name or f'HoldButton: {caption}'

	def on_click(self):
		""" HoldButton click event receiver. """

		self.hold_step = 0

	def on_hold(self):
		""" HoldButton hold event receiver. """

		if self.hold_step + 1 == self.hold_steps:
			super().on_click()
		else:
			super().on_hold()
		self.hold_step = min(self.hold_step + 1, self.hold_steps)

	def on_release(self):
		""" HoldButton release event receiver. """

		self.hold_step = 0
		super().on_release()

	def draw(self):
		""" HoldButton draw method. """

		try:
			if self.background_color is not None:
				self.app.fill_rect(self.x, self.y, self.width, self.height, color=self.background_color)
			self.app.draw_rect(self.x, self.y, self.width, self.height, line_width=self.app.button_border, color=self.color)
			progress = int((self.width - 4) * self.hold_step / max(1, self.hold_steps))
			progress = max(0, min(self.width - 4, progress))
			if progress > 0 and self.hold_color is not None:
				self.app.fill_rect(self.x + 2, self.y + 2, progress, self.height - 4, color=self.hold_color)
			x = int(self.x + self.width / 2)
			y = int(self.y + self.height / 2)
			self.app.draw_text(self.caption, x, y, color = self.color, font = self.font, alignment = 'center')
		except Exception as e:
			Log.error(e)

class CheckButton(Control):
	""" Check button user control that toggles self.checked. """

	def __init__(self, x: int|float, y: int|float, width: int|float, height: int|float,
	             box_color='white', check_color='red', check_width=5, background_color=None,
	             checked: bool=False, click=None, hold=None, release=None,
	             name: str=None, parameters: dict=None):
		super().__init__(x, y, width, height, click=click, hold=hold, release=release,
			name=name, parameters=parameters)
		self.checked = bool(checked)
		self.box_color = box_color
		self.check_color = check_color
		self.check_width = check_width
		self.background_color = background_color
		self.name = name or 'CheckButton'

	def on_click(self):
		""" CheckButton click event receiver. """

		self.checked = not self.checked
		super().on_click()

	def set_checked(self, checked: bool, do_click: bool=False):
		""" Programmatically set check state; optionally call the click handler
				if the check state changes. """

		if self.checked == checked:
			return
		if do_click is True:
			self.on_click()
		else:
			self.checked = checked

	def draw(self):
		""" CheckButton draw method."""

		try:
			if self.background_color is not None:
				self.app.fill_rect(self.x, self.y, self.width, self.height, color=self.background_color)
			self.app.draw_rect(self.x, self.y, self.width, self.height, line_width=self.app.button_border, color=self.box_color)
			if self.checked:		# draw x
				x1 = int(self.x + 3)
				y1 = int(self.y + 4)
				x2 = int(self.x + self.width - 6)
				y2 = int(self.y + self.height - 4)
				self.app.draw_line(x1, y1, x2, y2, width=self.check_width, color=self.check_color)
				self.app.draw_line(x1, y2, x2, y1, width=self.check_width, color=self.check_color)
		except Exception as e:
			Log.error(e)

class ToggleButton(Control):
	""" Toggle button user control that toggles self.toggled. """

	def __init__(self, x: int|float, y: int|float, width: int|float, height: int|float,
		toggled: bool=False, color='white', toggled_color='green', background_color='black',
		click=None, hold=None, release=None, name: str=None, parameters: dict=None):

		super().__init__(x, y, width, height, click=click, hold=hold, release=release,
			name=name, parameters=parameters)
		self.toggled = bool(toggled)
		self.color = color
		self.toggled_color = toggled_color
		self.background_color = background_color
		self.name = name or 'ToggleButton'

	def on_click(self):
		""" ToggleButton click event receiver. """

		self.toggled = not self.toggled
		super().on_click()

	def set_toggled(self, toggled: bool, do_click: bool=False):
		""" Programmatically set toggle state; optionally call the click handler
				if the toggle state changes."""

		if self.toggled == toggled:
			return
		if do_click is True:
			self.on_click()
		else:
			self.toggled = toggled

	def draw(self):
		""" ToggleButton draw method."""

		try:
			color = self.toggled_color if self.toggled else self.background_color
			if color is not None:
				self.app.fill_rect(self.x + 2, self.y + 2, self.width - 4, self.height - 4, color=color)
			self.app.draw_rect(self.x, self.y, self.width, self.height, line_width=self.app.button_border, color=self.color)
		except Exception as e:
			Log.error(e)

class SelectButton(Control):
	""" Select button control that cycles through a list of items.
			items can include a list of ints, strings, or None.
			If a values list is provided, then the items list provides the captions for the
			control, and .value is the entry in the .values list corresponding the selected item.
			Click advances to the next item, with wraparound.
			Current selection is available as .selected_index, .selected_item, and .value.
	"""

	def __init__(self, x: int|float, y: int|float, width: int|float, height: int|float,
		items: list, values: list=None, selected_index: int=0,
		color='white', background_color='black', font: Font|str=None,
		click=None, hold=None, release=None,
		name: str=None, parameters: dict=None):

		super().__init__(x, y, width, height, click=click, hold=hold, release=release,
			name=name, parameters=parameters)

		self.items = list(items) if items is not None else []
		self.values = values
		self.selected_index = None
		self.selected_item = None
		self.value = None
		self.color = color
		self.background_color = background_color
		self.font = font
		self.name = self.name or 'SelectButton'

		if self.values is not None and len(self.items) != len(self.values):
			Log.error(f'len(self.items) ({len(self.items)}) != len(self.values) ({len(self.values)})')
			return
		self.set_selected_index(selected_index)

	def set_items(self, items: list, values: list=None, reset_to_first: bool=True):
		""" Replace the items list; by default, reset selection to the first item. """

		self.items = list(items) if items is not None else []
		if len(self.items) == 0:
			self.selected_index = None
		elif reset_to_first or self.selected_item not in self.items or (self.selected_index is not None and self.selected_index >= len(self.items)):
			self.selected_index = 0
		elif self.selected_item in self.items:
			self.selected_index = self.items.index(self.selected_item)
		self.values = values if values is not None else self.values
		if self.values is not None and len(self.items) != len(self.values):
			Log.error(f'len(self.items) ({len(self.items)}) != len(self.values) ({len(self.values)})')
			self.selected_index = None
			self.selected_item = None
			self.value = None
			self.values = None
			return
		self.set_selected_index(self.selected_index)

	def on_click(self):
		""" SelectButton on_click event receiver. """

		if len(self.items) > 0:
			self.set_selected_index(0 if self.selected_index is None else (self.selected_index + 1) % len(self.items), do_click=True)

	def set_selected_item(self, item, do_click: bool=False):
		""" Programmatically set selection by item. Optionally calls the click handler
				if the selection changes. """

		if item not in self.items:
			Log.error(f'Item not in items: {item}')
			return
		index = self.items.index(item)
		self.set_selected_index(index, do_click=do_click)

	def set_selected_index(self, index: int, do_click: bool=False):
		""" Programmatically set selection by index. Optionally calls the click handler
				if the selection changes. """

		if index is None:
			self.selected_index = None
			self.selected_item = None
			self.value = None
			return
		if len(self.items) == 0:
			Log.error('No items to select')
			return
		if index < 0 or index >= len(self.items):
			Log.error(f'Index out of range: {index}')
			return
		if self.selected_index == index:
			return
		self.selected_index = index
		self.selected_item = self.items[index]
		self.value = self.values[self.selected_index] if self.values is not None else None
		if do_click is True:
			super().on_click()

	def draw(self):
		""" SelectButton draw method. """

		try:
			if self.background_color is not None:
				self.app.fill_rect(self.x, self.y, self.width, self.height,
					color=self.background_color)
			self.app.draw_rect(self.x, self.y, self.width, self.height,
				line_width=self.app.button_border, color=self.color)
			if self.selected_index is None:
				return
			label = str(self.selected_item).strip()
			if len(label) == 0:
				return
			cx = int(self.x + self.width / 2)
			cy = int(self.y + self.height / 2)
			self.app.draw_text(label, cx, cy, color=self.color, font=self.font, alignment='center')
		except Exception as e:
			Log.error(e)



""" Jubilee Log mode class. """

import math, platform, time
try:
	import psutil
except:
	psutil = None
from .mode import Mode
from .controls import Button
from .misc import Log

class LogMode(Mode):
	""" Log mode class. """

	def __init__(self, background_color: str='black'):
		self.log_date = None
		self.log_text = []
		self.log_line_height = 11
		self.log_lines_per_page = 25
		self.date_width = 75
		self.font_size = 12
		self.class_name_width = 125
		self.method_name_width = 125
		self.level_name_width = 40
		self.log_page = 0
		self.max_log_size = 100
		self.button_font = None
		self.return_mode = None
		self.return_mode_parameters = None
		self.back_button = None
		self.last_process = None
		self.cpu_load = []
		self.cpu_temp = []
		super().__init__(background_color=background_color)

	def init(self):
		""" LogMode initializer. """

		self.name = 'Log'

		# add page size from app
		self.log_lines_per_page = int((self.app.screen_height - 180) / self.log_line_height)

		# create controls for log mode
		button_width = 77
		self.add_control(Button('Font', self.app.button_margin, self.app.screen_height - 60,
			button_width, 60, font=self.button_font, click=self.change_font))
		self.add_control(Button('Up', button_width + self.app.button_margin * 2,
			self.app.screen_height - 60, button_width, 60, font=self.button_font,
			click=self.log_page_up))
		self.add_control(Button('Down', button_width * 2 + self.app.button_margin * 3,
			self.app.screen_height - 60, button_width, 60, font=self.button_font,
			click=self.log_page_down))

		# create Back button
		self.back_button = Button('Back', self.app.screen_width-77-self.app.button_margin,
			self.app.screen_height-60, 77, 60, font=self.button_font,
			click=self.back_click)
		self.add_control(self.back_button)

		# By default, the Back button causes the app to return to the mode from which it
		# was called. The return mode can be set statically in this initializer, or
		# dynamically using this code:
		# 	self.app.modes['Log'].back_button.target_mode = return_mode

	def enter(self, mode_parameters: dict=None):
		""" Log mode enter method. """

		super().enter(mode_parameters)
		self.return_mode = mode_parameters.get('previous_mode', self.return_mode)
		self.log_page = 0
		self.check_log()

	def back_click(self):
		""" Handles Back button click. """

		mode = self.back_button.target_mode or self.return_mode
		if mode is None:
			Log.error('No mode to switch back to')
		else:
			self.app.set_mode(mode, mode_parameters=self.return_mode_parameters)

	def process(self):
		""" Log mode process method. Runs at 1 Hz. """

		try:
			now = time.time()
			if self.last_process is not None and now - self.last_process < 1:
				return
			self.last_process = now
			self.check_log()
			self.record_cpu_temperatures()
		except Exception as e:
			Log.error(e)

	def check_log(self):
		""" Check status of log and reload if changed. """

		try:
			modification_date = Log.get_modification_date()
			if modification_date == self.log_date:
				return
			self.log_date = modification_date
			log_lines = Log.read()[-self.max_log_size:]
			self.log_text = list(reversed(log_lines))
		except Exception as e:
			Log.error(e)

	def record_cpu_temperatures(self):
		""" Record CPU temperatures. """

		if psutil is None:
			return
		try:
			max_graph_points = self.app.screen_width - 180
			try:
				self.cpu_load.append(int(psutil.cpu_percent()))
				self.cpu_load = self.cpu_load[-max_graph_points:]
			except:
				pass
			temperature = None
			if platform.system() != 'Darwin':
				temperature_metrics = psutil.sensors_temperatures() or {}
				for key in ('cpu-thermal', 'cpu_thermal', 'coretemp', 'k10temp', 'soc_thermal'):
					if temperature_metrics.get(key) is not None:
						temperature = int(temperature_metrics[key][0].current)
						break
			if temperature is not None:
				self.cpu_temp.append(temperature)
				self.cpu_temp = self.cpu_temp[-max_graph_points:]
		except Exception as e:
			Log.error(e)

	def draw(self):
		""" Log mode draw method. """

		try:
			# draw CPU load
			graph_x = 90
			if len(self.cpu_load) > 0:
				self.app.draw_text(f'CPU: {self.cpu_load[-1]}%', self.app.margin, 20)
				self.app.draw_line(graph_x, 10, graph_x, 49)
				self.app.draw_line(graph_x, 49, self.app.screen_width - 20, 49)
				for i in range(1, len(self.cpu_load)):
					cpu = 48 - int(self.cpu_load[i] / 100.0 * 38.0)
					self.app.draw_pixel(graph_x + i + 1, cpu)

			# draw CPU temp
			if len(self.cpu_temp) > 0:
				self.app.draw_text(f'Temp: {self.cpu_temp[-1]} C', self.app.margin, 70)
				self.app.draw_line(graph_x, 60, graph_x, 99)
				self.app.draw_line(graph_x, 99, self.app.screen_width - 20, 99)
				for i, temp in enumerate(self.cpu_temp):
					cpu = 98 - int(temp / 100.0 * 38.0)
					self.app.draw_pixel(graph_x + i + 1, cpu)

			# check page_down and ensure that it has not gone past end of log
			if self.log_lines_per_page < 1:
				Log.debug(f'log_lines_per_page = {self.log_lines_per_page}')
				return
			num_pages = math.ceil(len(self.log_text) / self.log_lines_per_page)
			self.log_page = max(0, min(self.log_page, num_pages - 1))

			# draw log
			y = 95 + self.log_line_height
			self.app.draw_text('Log', self.app.margin, y)
			y += self.app.underscore_position
			self.app.draw_line(self.app.margin, y, self.app.screen_width - self.app.margin, y)
			font = self.app.standard_font_sizes[self.font_size]
			for i in range(self.log_lines_per_page):
				line_number = self.log_page * self.log_lines_per_page + i
				if line_number >= len(self.log_text):
					break
				record_fields = Log.parse(self.log_text[line_number])
				if record_fields is None:
					continue
				y += self.log_line_height
				self.app.draw_text(record_fields['dt'].strftime('%Y%m%d %H:%M:%S'), self.app.margin, y, font=font)
				x = self.app.margin + self.date_width
				self.app.draw_text(record_fields['class'], x, y, font=font)
				x += self.class_name_width
				self.app.draw_text(record_fields['method'], x, y, font=font)
				x += self.method_name_width
				self.app.draw_text(record_fields['level'], x, y, font=font)
				x += self.level_name_width
				self.app.draw_text(record_fields['message'], x, y, font=font)
		except Exception as e:
			Log.error(e)

	def change_font(self):
		""" Changes font. """

		self.app.change_font()
		self.app.set_popover(f'Changed font to {self.app.standard_font}')

	def log_page_up(self):
		""" Scrolls log up one page. """

		self.log_page = max(self.log_page - 1, 0)

	def log_page_down(self):
		""" Scrolls log down one page. """

		self.log_page = self.log_page + 1



""" Jubilee Shut Down mode class. """

from .mode import Mode
from .controls import Button, HoldButton
from .misc import Log

class ShutDownMode(Mode):
	""" Shutdown mode class. """

	def __init__(self):
		super().__init__()
		self.return_mode = None

	def init(self):
		""" Shut Down mode initializer. """

		self.name = 'Shut Down'
		button_width = 150
		hold_steps = int(self.app.config.get('app_process_fps', 10) * 0.75)
		self.add_control(HoldButton('Yes', self.app.button_margin, self.app.screen_height - 60, button_width, 60, click=self.app.shut_down, hold_color='red', hold_steps=hold_steps))
		self.add_control(Button('Cancel', self.app.screen_width - self.app.button_margin - button_width, self.app.screen_height - 60, button_width, 60, click=self.cancel_shutdown))

	def enter(self, mode_parameters: dict=None):
		""" Shut Down mode enter method. """

		self.return_mode = (mode_parameters or {}).get('previous_mode', None)

	def draw(self):
		""" Shut Down mode draw method. """

		try:
			self.app.center_text('Confirm Shutdown', self.app.screen_middle - 30)
			self.app.center_text('Hold to confirm')
		except Exception as e:
			Log.error(e)

	def cancel_shutdown(self):
		""" Cancel shutdown and return to previous mode. """

		if self.return_mode is None:
			Log.error('return_mode is None')
		else:
			self.app.set_mode(self.return_mode)



""" Jubilee base classes. """

import enum
from pygame import Surface
from pygame.event import Event
from .misc import Log

class Animation:
	""" Animation class. Stores a set of frames and a set of sequences. """

	def __init__(self, name: str=None, frames: list=None, sequences: dict=None):
		self.name = name
		self.frames = frames or []
		self.sequences = sequences or {}			# {'sequence name': [frame numbers]}

class SpritePosition(enum.Enum):
	""" Sprite positions enum. """

	TopLeft = 1															# x/y maps to top-left
	Center = 2															# x/y maps to center
	Bottom = 3															# x/y maps to bottom

class Sprite:
	""" Sprite class. """

	def __init__(self, name: str=None, static_image: str=None, animation: str|Animation=None, auto_animate_rate: int=None):
		self.app = None
		self.name = name
		self.x = None
		self.y = None
		self.z = None											# Z-order overrides positioning; None = layer 255
		self.static_image = static_image	# set to Surface or image name for static image
		self.animation = animation				# set to an Animation or animation name
		self.sequence_name = None					# set to a sequence name for animation
		self.sequence_frame = None
		self.auto_animate_rate = max(1, int(auto_animate_rate) if auto_animate_rate is not None else 0)
		self.auto_animate_step = 0
		self.scale = None									# set to either a float or a two-tuple
		self.flip_x = False								# flip horizontally
		self.flip_y = False								# flip vertically
		self.rotate = None								# set to a float
		self.hue_shift = None							# set to hue delta, range 0-360
		self.image = None									# this is set by set_image()
		self.width = None									# this is set by set_image()
		self.height = None								# this is set by set_image()

	def bind(self, app):
		""" Binds sprite to app. """

		self.app = app
		# try to find animation from string passed as animation name
		if self.animation is not None and isinstance(self.animation, str):
			self.animation = self.app.get_animation(self.animation)
		# if animation is still None, try to find animation from sprite name
		if self.animation is None and self.name is not None:
			self.animation = self.app.get_animation(self.name)

	def set_static_image(self, image: str|Surface|None):
		""" Sets self.static_image based on image lookup. """

		self.static_image = self.app.get_image(image)

	def set_image(self) -> Surface|None:
		""" Sets current image for sprite based on self.static_image or
				self.sequence_name/self.sequence_frame. Also applies transformations
				and calls set_size(). """

		try:
			self.image = None
			if self.static_image is not None:
				self.image = self.app.get_image(self.static_image)
			else:
				animation = self.app.get_animation(self.animation)
				if animation is None:
					return None
				if self.sequence_name is None:
					if self.sequence_frame is None:
						self.sequence_frame = 0
					if self.sequence_frame >= 0 and self.sequence_frame < len(animation.frames):
						self.image = animation.frames[self.sequence_frame]
				else:
					if self.sequence_name not in animation.sequences:
						return None
					sequence = animation.sequences.get(self.sequence_name)
					if self.sequence_frame is None:
						self.sequence_frame = 0
					if self.sequence_frame < 0 or self.sequence_frame >= len(sequence):
						return None
					sequence_frame = sequence[self.sequence_frame]
					if sequence_frame >= 0 and sequence_frame < len(animation.frames):
						self.image = animation.frames[sequence_frame]

			if self.image is None:
				return None

			# apply transformations
			if self.scale is not None:
				if isinstance(self.scale, float):
					self.image = self.app.scale_image(self.image, self.scale)
				elif isinstance(self.scale, (tuple, list)) and len(self.scale) == 2:
					self.image = self.app.scale_image(self.image, self.scale[0], self.scale[1])
			if self.flip_x is True or self.flip_y is True:
				self.image = self.app.flip_image(self.image, horizontal=self.flip_x, vertical=self.flip_y)
			if self.rotate is not None:
				self.image = self.app.rotate_image(self.image, self.rotate)
			if self.hue_shift is not None:
				self.image = self.app.shift_image_hue(self.image, self.hue_shift)

			# set size and return image
			self.width, self.height = self.image.get_size()
			return self.image

		except Exception as e:
			Log.error(e)
			return None

	def set_sequence(self, sequence_name: str, auto_animate_rate: int=None, reset_sequence_frame: bool=True) -> bool:
		""" Sets an animation sequence, optionally with an animation rate. """

		try:
			animation = self.app.get_animation(self.animation)
			if animation is None:
				Log.error('No animation to set')
				return False
			if sequence_name not in animation.sequences:
				Log.error(f'No sequence named {sequence_name} in animation')
				self.sequence_name = None
				return False
			self.sequence_name = sequence_name
			self.auto_animate_rate = 0 if auto_animate_rate is None else max(1, int(self.auto_animate_rate))
			if reset_sequence_frame is True or self.sequence_frame >= len(animation.sequences[sequence_name]):
				self.sequence_frame = None
				self.auto_animate_step = 0
			return True
		except Exception as e:
			Log.error(e)
			return False

	def auto_animate(self):
		""" Performs auto-animation. """

		if self.static_image is not None or self.animation is None or self.auto_animate_rate == 0:
			return
		self.auto_animate_step = self.auto_animate_step + 1
		if self.auto_animate_step >= self.auto_animate_rate:
			self.auto_animate_step = 0
			self.animate()

	def animate(self, sequence_frame: int=None) -> bool:
		""" Advances animation to the next frame in the sequence. """

		try:
			if self.static_image is not None:
				Log.debug(f'Not animating sprite {self.name} because set to static image')
				return True

			# find animation
			animation = self.app.get_animation(self.animation)
			if animation is None:
				Log.error('No animation to set')
				return False
			if len(animation.frames) == 0:
				Log.error(f'Animation {animation.name} has no frames')
				return False

			if self.sequence_name is None:		# animate through all frames
				if sequence_frame is not None:
					self.sequence_frame = sequence_frame
				else:
					self.sequence_frame = 0 if self.sequence_frame is None else (self.sequence_frame + 1) % len(animation.frames)

			else:	# animate based on specified sequence
				sequence = animation.sequences.get(self.sequence_name)
				if sequence is None:
					Log.error(f'No sequence named {self.sequence_name}')
					return False
				if len(sequence) == 0:
					Log.error(f'Sequence {self.sequence_name} has no frames')
					return False
				if sequence_frame is not None:
					self.sequence_frame = sequence_frame
				else:
					self.sequence_frame = 0 if self.sequence_frame is None else (self.sequence_frame + 1) % len(sequence)
				if self.sequence_frame is None or self.sequence_frame < 0 or self.sequence_frame >= len(sequence) or sequence[self.sequence_frame] >= len(animation.frames):
					Log.error(f'Invalid frame number {self.sequence_frame} for animation {animation.name} and sequence {self.sequence_name}')
					return False

			return True

		except Exception as e:
			Log.error(e)
			return False

	def process(self):
		""" Sprite process stub method. """

class PointerInterface:
	""" Jubilee pointer interface class - base class for MouseInterface and TouchInterface. """

	def __init__(self):
		self.x = None
		self.y = None
		self.down = False
		self.held = False

	def handle_event(self, event: Event):
		""" Event handler function for events. """

	def detect_events(self):
		""" Event detector function for polled devices. """

	def release(self):
		""" Resource release function. """



""" Jubilee Worker class. """

import datetime, json, multiprocessing, os, queue, sys, time
import __main__
from .misc import Config, Log

class Worker:
	""" Worker class. """

	config_defaults = {'headless': False,
		'nosound': False,
		'screen_resolution': [320, 240],
		'screen_scale': [[0, 319, -1], [0, 239, 1]],
		'screen_rotation': 0,
		'app_process_fps': 20,
		'app_draw_fps': 20, 'modal': True,
		'worker_process_fps': 20,
		'worker_process_periodic_fps': 1,
		'persist_app_state': True,
		'app_state_filename': 'app_state.txt',
		'app_state_start_filename': 'app_state_start.txt',
		'keyboard_input': True,
		'log_rotation': 'daily',
		'font_size': 14
	}

	def __init__(self, app_queue, worker_queue, config_manager=False, log_manager=False):
		self.name = 'Worker'
		self.app_queue = app_queue
		self.worker_queue = worker_queue
		self.config_manager = config_manager
		self.log_manager = log_manager
		self.worker_process = None
		self.base_path = os.path.dirname(os.path.realpath(__main__.__file__))
		self.config_filename = os.path.join(self.base_path, 'config.txt')
		self.config = Config.load(self.config_filename, defaults=self.config_defaults)
		self.config_date = None
		self.log_date = None
		if os.path.isfile(self.config_filename) is True:
			self.config_date = os.path.getmtime(self.config_filename)
		self.last_periodic = None
		self.init()

		# start process
		self.worker_process = multiprocessing.Process(target=self.run)
		self.worker_process.daemon = True
		self.worker_process.start()

	def init(self):
		""" Worker-specific initializer that runs before the worker starts.
					State changes made here are visible to the App.
		"""

	def start_worker(self):
		""" Worker-specific initializer that runs at the start of run loop.
					State changes made here are *not* visible to the App.
		"""

	def run(self):
		""" Worker run loop. """

		try:
			Log.info('Starting')
			self.start_worker()

			while True:
				loop_start = time.time()
				self.receive_messages()

				# call main process function
				self.on_process()

				# call periodic process function occasionally
				process_periodic_fps = self.config.get('worker_process_periodic_fps', 1)
				if process_periodic_fps is not None:
					elapsed = time.time() - (self.last_periodic or 0)
					if elapsed >= 1 / process_periodic_fps:
						self.on_process_periodic()

				# delay to next loop
				loop_time = time.time() - loop_start
				delay = 1 / max(1, int(self.config.get('worker_process_fps', 10))) - loop_time
				if delay > 0:
					time.sleep(delay)

		except Exception as e:
			Log.error(e)

	def manage_config(self):
		""" Config manager. """

		if self.config_manager is False:
			return
		if os.path.isfile(self.config_filename) is False:
			self.config_date = None
			return
		config_date = os.path.getmtime(self.config_filename)
		if config_date != self.config_date:
			Log.info(f'Loading config ({self.config_date} != {config_date})')
			self.config_date = config_date
			self.config = Config.load(self.config_filename, defaults=self.config_defaults)
			self.send_updated_config()

	def manage_log(self):
		""" Log manager. """

		if self.log_manager is False:
			return
		if os.path.isfile(Log.get_filename()) is False:
			self.log_date = None
			return
		self.log_date = self.log_date or datetime.datetime.now()
		rotation = self.config.get('log_rotation', 'daily')
		rotate = False
		log_format = None
		if rotation == 'daily':
			rotate = (datetime.datetime.now().strftime('%Y%m%d') != self.log_date.strftime('%Y%m%d'))
			log_format = '%Y%m%d'
		if rotation == 'monthly':
			rotate = (datetime.datetime.now().strftime('%Y%m') != self.log_date.strftime('%Y%m'))
			log_format = '%Y%m'
		if rotation == 'hourly':
			rotate = (datetime.datetime.now().strftime('%Y%m') != self.log_date.strftime('%Y%m'))
			log_format = '%Y%m%d%H'
		if rotate is True:
			filename = f'log_{datetime.datetime.now().strftime(log_format)}.txt'
			Log.backup(backup_filename=filename)
			self.log_date = datetime.datetime.now()

	def on_process(self):
		""" Regular (high-frequency) worker processing event receiver.
				This merely wraps process() but is a placeholder for future
				functionality and for consistency with other event receivers. """

		self.process()

	def process(self):
		""" Regular (high-frequency) worker processing stub function. """

	def on_process_periodic(self):
		""" Periodic (low-frequency) worker processing event receiver. """

		process_periodic_fps = self.config.get('worker_process_periodic_fps', 1)
		self.last_periodic = (self.last_periodic or time.time()) + 1.0 / process_periodic_fps
		if self.config_manager is True:
			self.manage_config()
		if self.log_manager is True:
			self.manage_log()
		self.process_periodic()

	def process_periodic(self):
		""" Periodic (low-frequency) worker processing stub function. """

	def exit(self, code=0):
		sys.exit(code)

	# messaging with app

	def send_message(self, message: str|dict):
		""" Send a message to the app. """

		if isinstance(message, str):
			message = {'action': message}
		try:
			self.worker_queue.put(json.dumps(message, ensure_ascii=True))
		except Exception as e:
			Log.error(f'Failed to send message: {e}')

	def receive_messages(self):
		""" Receive messages from app. """

		while True:
			try:
				message = self.app_queue.get_nowait()
				self.process_message(json.loads(message), sender='App')
			except queue.Empty:
				return
			except Exception as e:
				Log.error(e)
				continue

	def process_message(self, message: dict, sender: str=None):
		""" Process a message from app. This method can be extended in subclass. """

		action = message.get('action')
		if action == 'update config':
			key = message.get('key')
			value = message.get('value')
			self.update_config(key, value)
		elif action == 'config updated':
			self.config = message.get('config', {})
		elif action == 'exit':
			self.exit()
		else:
			Log.warning(f'Received unknown message: {message}')

	def update_config(self, key, value):
		self.config[key] = value
		self.write_config()

	def write_config(self):
		Config.save(self.config, self.config_filename)
		self.config_date = os.path.getmtime(self.config_filename) if os.path.isfile(self.config_filename) else None
		self.send_updated_config()

	def send_updated_config(self):
		message = {'action': 'config updated', 'config': self.config}
		self.send_message(message)



""" Jubilee Mode class. """

import inspect, os
from .base_classes import Sprite, SpritePosition
from .controls import Control
from .misc import Log

class Mode:
	""" Jubilee Mode class. """

	def __init__(self, background_color: str='black'):

		# app and mode settings
		self.app = None
		if hasattr(self, 'name') is False:		# preserve name declared in subclass
			self.name = 'Unnamed Mode'
		self.background_color = background_color
		self.mode_timer = None
		self.submode = None
		self.submode_timer = None

		# find submodes by introspection
		method_names = list(m[0] for m in inspect.getmembers(self, predicate=inspect.ismethod))
		self.submodes = []
		for method_type in ['enter', 'click', 'process', 'draw', 'exit']:
			for m in (m for m in method_names if m.startswith(f'{method_type}_')):
				submode_name = m[len(method_type) + 1:]
				if len(submode_name) > 0 and submode_name not in self.submodes:
					self.submodes.append(submode_name)

		# controls
		self.controls = []		# Z-ordered from highest to lowest
		self.show_controls = True
		self.selected_control = None

		# load resources
		self.images_path = None
		self.images = {}
		self.animations = {}
		self.sounds_path = None
		self.sounds = {}
		self.sprites = []
		self.sprite_positions = SpritePosition.Bottom		# topleft, center, bottom, or None

	def init(self):
		""" Mode-specific stub initializer. """

	def load_resources(self):
		""" Mode resource loading. Called during app.add_mode() after init(). """

		self.images_path = os.path.join(self.app.base_path, self.name, 'images')
		self.images, self.animations = self.app.load_images(self.images_path)
		self.sounds_path = os.path.join(self.app.base_path, self.name, 'sounds')
		self.sounds = self.app.load_sounds(self.sounds_path)

	def on_enter(self, mode_parameters: dict=None):
		""" Mode enter event receiver. """

		self.mode_timer = 0
		try:
			self.enter(mode_parameters=mode_parameters)
		except Exception as e:
			Log.error(e)

	def enter(self, mode_parameters: dict=None):
		""" Mode enter event stub method. """

	def set_submode(self, name: str|None, mode_parameters: dict=None):
		""" Sets submode and resets submode timer. """

		# call exit_submode on current submode if it exists
		if self.submode is not None and hasattr(self, f'exit_{self.submode}'):
			try:
				getattr(self, f'exit_{self.submode}')()
			except Exception as e:
				Log.error(f'Exception exiting submode {self.submode}: {e}')
		self.submode = None
		self.submode_timer = None

		if name not in self.submodes:
			if name is not None:
				Log.error(f'No known submode {name}')
			return

		self.submode = name
		self.submode_timer = 0
		if hasattr(self, f'enter_{name}'):
			try:
				getattr(self, f'enter_{name}')(mode_parameters)
			except Exception as e:
				Log.error(f'Exception entering submode {self.submode}: {e}')

	def add_control(self, control) -> Control:
		""" Add control to mode. """

		control.bind(self.app)
		self.controls.append(control)
		return control

	def get_control(self, name: str) -> Control|None:
		""" Gets control by name. If several controls have the name, returns the first one. """

		matching_controls = list(c for c in self.controls if c.name == name)
		return matching_controls[0] if len(matching_controls) > 0 else None

	def remove_control(self, control):
		""" Remove control from mode. Can either pass in the control or its caption. """

		if isinstance(control, str):
			matching_controls = list(c for c in self.controls if c.caption == control)
			if len(matching_controls) > 0:
				for b in matching_controls:
					self.controls.remove(b)
			else:
				Log.error(f'No control with caption {control} in mode')
		else:
			if control in self.controls:
				self.controls.remove(control)
			else:
				Log.error(f'Control {control} is not in mode.controls')

	def remove_controls(self):
		""" Remove all controls from mode. """

		self.controls = []

	def on_click(self, x: int|float, y: int|float):
		""" Mode click event receiver. """

		# test controls first
		if self.show_controls is True:
			for control in self.controls:
				if control.visible is True and control.collide(x, y):
					if control.enabled is True:
						self.selected_control = control
						try:
							control.on_click()
						except Exception as e:
							Log.error(f'Exception clicking control {control.name}: {e}')
					return

		# send to click mode or submode
		if self.submode is not None and hasattr(self, f'click_{self.submode}'):
			try:
				getattr(self, f'click_{self.submode}')(x, y)
			except Exception as e:
				Log.error(f'Exception clicking mode {self.name} submode {self.submode}: {e}')
		else:
			try:
				self.click(x=x, y=y)
			except Exception as e:
				Log.error(e)

	def click(self, x: int|float, y: int|float):
		""" Mode click event stub method. """

	def on_hold(self):
		""" Mode hold event receiver. """

		if self.selected_control is not None:
			try:
				self.selected_control.on_hold()
			except Exception as e:
				Log.error(f'Exception holding control {self.selected_control.name}: {e}')
		else:
			try:
				self.hold()
			except Exception as e:
				Log.error(e)

	def hold(self):
		""" Mode hold event stub method. """

	def on_release(self):
		""" Mode release event receiver. """

		if self.selected_control is not None:
			try:
				self.selected_control.on_release()
			except Exception as e:
				Log.error(f'Exception releasing control {self.selected_control.name}: {e}')
			self.selected_control = None
		else:
			try:
				self.release()
			except Exception as e:
				Log.error(e)

	def release(self):
		""" Mode release event stub method. """

	def on_process(self):
		""" Mode process event receiver. """

		if self.submode is not None and hasattr(self, f'process_{self.submode}'):
			try:
				try:
					getattr(self, f'process_{self.submode}')()
				except Exception as e:
					Log.error(f'Error processing mode {self.name} submode {self.submode}: {e}')
			except Exception as e:
				Log.error(f'Exception processing mode {self.name} submode {self.submode}: {e}')
		else:
			try:
				self.process()
			except Exception as e:
				Log.error(f'Error processing mode {self.name}: {e}')
		for sprite in self.sprites:
			sprite.process()

	def process(self):
		""" Mode process stub method. """

	def on_draw(self):
		""" Mode draw event receiver. """

		self.mode_timer += 1
		if self.background_color is not None:
			self.app.fill_screen(self.background_color)
		if self.submode is not None and hasattr(self, f'draw_{self.submode}'):
				self.submode_timer += 1
				try:
					getattr(self, f'draw_{self.submode}')()
				except Exception as e:
					Log.error(f'Exception drawing mode {self.name} submode {self.submode}: {e}')
		else:
			try:
				self.draw()
			except Exception as e:
				Log.error(e)

		# draw controls in reverse order, i.e., back-to-front if overlapping
		if self.show_controls is True:
			for control in reversed(self.controls):
				if control.visible is True:
					try:
						control.draw()
					except Exception as e:
						Log.error(f'Exception drawing mode {self.name} control {control.name}: {e}')

	def draw(self):
		""" Mode draw stub method. """

	def add_sprite(self, sprite) -> Sprite:
		""" Adds sprite. """

		sprite.bind(self.app)
		self.sprites.append(sprite)
		return sprite

	def get_sprite(self, name: str) -> Sprite|None:
		""" Gets sprite by name. If several sprites have the name, returns the first one. """

		matching_sprites = list(s for s in self.sprites if s.name == name)
		return matching_sprites[0] if len(matching_sprites) > 0 else None

	def render_sprites(self, auto_animate: bool=True):
		""" Draws sprites on window, optionally calling auto_animate on each.
				Sprites are drawn in the order defined by sprite_positions: topleft, center,
				bottom, or None. """

		# first auto-animate all sprites to set sizes
		for s in self.sprites:
			if auto_animate:
				s.auto_animate()
			s.set_image()

		# sort by sprite positions, with Z-order taking priority
		if self.sprite_positions is not None:
			self.sprites.sort(key=lambda s: (s.y or 0) * self.app.screen_width + (s.x or 0))
		if any(s.z is not None for s in self.sprites):
			self.sprites.sort(key=lambda s: 100 if s.z is None else s.z, reverse=True)

		# render
		for s in self.sprites:
			if s.image is not None:
				self.app.blit(s.image, s.x, s.y, position=self.sprite_positions)

	def remove_sprite(self, sprite):
		""" Removes sprite. """

		if sprite not in self.sprites:
			Log.error('Sprite is not in mode.sprites')
			return
		self.sprites.remove(sprite)

	def remove_sprites(self):
		""" Removes all sprites. """

		self.sprites = []

	def on_exit(self):
		""" Mode exit event receiver. """

		self.mode_timer = None

		if self.selected_control is not None:
			try:
				self.selected_control.on_release()
			except Exception as e:
				Log.error(f'Exception calling on_release() on control {self.selected_control.name}: {e}')
			self.selected_control = None

		# call exit_submode on current submode if it exists, and set submode to None
		if self.submode is not None and hasattr(self, f'exit_{self.submode}'):
				try:
					getattr(self, f'exit_{self.submode}')()
				except Exception as e:
					Log.error(f'Exception exiting mode {self.name} submode {self.submode}: {e}')
		try:
			self.exit()
		except Exception as e:
			Log.error(e)

		self.submode = None

	def exit(self):
		""" Mode exit event stub method. """
